from cached_property import cached_property
from onegov.core.templates import render_macro
from onegov.directory import Directory, DirectoryEntry
from onegov.event import EventCollection
from onegov.form import FormSubmissionCollection
from onegov.org import _
from onegov.org.layout import DefaultLayout, EventLayout
from onegov.chat import Message
from onegov.core.elements import Link, LinkGroup, Confirm, Intercooler
from onegov.reservation import Allocation, Resource, Reservation
from onegov.ticket import Ticket, Handler, handlers
from onegov.search.utils import extract_hashtags
from purl import URL
from sqlalchemy.orm import object_session


class OrgTicketMixin(object):
    """ Adds additional methods to the ticket, needed by the organisations
    implementation of it. Strictly limited to things that
    do not belong into onegov.ticket.

    """

    def reference(self, request):
        """ Returns the reference which should be used wherever we talk about
        a ticket, but do not show it (say in an e-mail subject).

        This reference should not be changed so it stays consistent.

        If you want, you can override the content of the reference group,
        shown in brackets (see :meth:`reference_group`).

        """
        return f'{self.number} / {self.reference_group(request)}'

    def reference_group(self, request):
        return request.translate(self.group)

    @cached_property
    def extra_localized_text(self):

        # extracts of attachments are currently not searchable - if they were
        # we would add this here - probably in a raw SQL query that
        # concatenates all the text
        #
        # for now I've decided against it as it would lower the hit-rate
        # for notes (which should be very easy to find), just to be able
        # to search through files which are mostly going to be irrelevant
        # for what the user wants to find
        #
        # if the user wants to have a ticket findable through the file content
        # we should advise them to enter a meaningful note with the file
        # instead.
        #
        q = object_session(self).query(Message)
        q = q.filter_by(channel_id=self.number)
        q = q.filter(Message.type.in_(('ticket_note', 'ticket_chat')))
        q = q.with_entities(Message.text)

        return ' '.join(n.text for n in q if n.text)

    @property
    def es_tags(self):
        if self.extra_localized_text:
            return [
                tag.lstrip('#') for tag in extract_hashtags(
                    self.extra_localized_text
                )
            ]


class FormSubmissionTicket(OrgTicketMixin, Ticket):
    __mapper_args__ = {'polymorphic_identity': 'FRM'}
    es_type_name = 'submission_tickets'


class ReservationTicket(OrgTicketMixin, Ticket):
    __mapper_args__ = {'polymorphic_identity': 'RSV'}
    es_type_name = 'reservation_tickets'


class EventSubmissionTicket(OrgTicketMixin, Ticket):
    __mapper_args__ = {'polymorphic_identity': 'EVN'}
    es_type_name = 'event_tickets'

    def reference_group(self, request):
        return self.title


class DirectoryEntryTicket(OrgTicketMixin, Ticket):
    __mapper_args__ = {'polymorphic_identity': 'DIR'}
    es_type_name = 'directory_tickets'


@handlers.registered_handler('FRM')
class FormSubmissionHandler(Handler):

    handler_title = _("Form Submissions")

    @cached_property
    def collection(self):
        return FormSubmissionCollection(self.session)

    @cached_property
    def submission(self):
        return self.collection.by_id(self.id)

    @cached_property
    def form(self):
        return self.submission.form_class(data=self.submission.data)

    @property
    def deleted(self):
        return self.submission is None

    @property
    def email(self):
        return self.submission.email

    @property
    def title(self):
        return self.submission.title

    @property
    def subtitle(self):
        return None

    @property
    def group(self):
        return self.submission.form.title

    @property
    def payment(self):
        return self.submission and self.submission.payment

    @property
    def extra_data(self):
        return self.submission and [
            v for v in self.submission.data.values()
            if isinstance(v, str)
        ]

    @property
    def undecided(self):

        # submissions without registration window do not present a decision
        if not self.submission.registration_window:
            return False

        if self.submission.claimed is None:
            return True

        return False

    def get_summary(self, request):
        layout = DefaultLayout(self.submission, request)

        return render_macro(layout.macros['display_form'], request, {
            'form': self.form,
            'layout': layout
        })

    def get_links(self, request):
        layout = DefaultLayout(self.submission, request)

        links = []
        extra = []

        # there's a decision to be made about the registration
        window = self.submission.registration_window

        if window:
            if self.submission.spots and self.submission.claimed is None:
                confirmation_traits = [
                    Intercooler(
                        request_method='POST',
                        redirect_after=request.url
                    )
                ]

                next_in_queue = window.next_submission

                if next_in_queue and next_in_queue is not self.submission:
                    confirmation_traits.append(Confirm(
                        _(
                            "This is not the oldest undecided submission of "
                            "this registration window. Do you really want to "
                            "confirm this submission?"
                        ),
                        _(
                            "By confirming this submission, you will prefer "
                            "this over a submission that came in earlier."
                        ),
                        _(
                            "Confirm registration"
                        ),
                        _(
                            "Cancel"
                        )
                    ))

                links.append(
                    Link(
                        text=_("Confirm registration"),
                        url=request.return_here(
                            layout.csrf_protected_url(
                                request.link(
                                    self.submission, 'confirm-registration')
                            )
                        ),
                        attrs={'class': 'accept-link'},
                        traits=confirmation_traits
                    )
                )
                extra.append(
                    Link(
                        text=_("Deny registration"),
                        url=request.return_here(
                            layout.csrf_protected_url(
                                request.link(
                                    self.submission, 'deny-registration')
                            )
                        ),
                        attrs={'class': 'delete-link'},
                        traits=(
                            Intercooler(
                                request_method='POST',
                                redirect_after=request.url
                            ),
                        )
                    )
                )

            # a registration was accepted before, we can issue an uninvite
            if self.submission.spots and self.submission.claimed:
                links.append(
                    Link(
                        text=_("Cancel registration"),
                        url=request.return_here(
                            layout.csrf_protected_url(
                                request.link(
                                    self.submission, 'cancel-registration')
                            )
                        ),
                        attrs={'class': 'delete-link'},
                        traits=(
                            Intercooler(
                                request_method='POST',
                                redirect_after=request.url
                            ),
                        )
                    )
                )

        edit_link = URL(request.link(self.submission))
        edit_link = edit_link.query_param('edit', '').as_string()

        (links if not links else extra).append(
            Link(
                text=_('Edit submission'),
                url=request.return_here(edit_link),
                attrs={'class': 'edit-link'}
            )
        )

        if extra:
            links.append(LinkGroup(
                _("Advanced"),
                links=extra,
                right_side=False
            ))

        return links


@handlers.registered_handler('RSV')
class ReservationHandler(Handler):

    handler_title = _("Reservations")

    @cached_property
    def resource(self):
        query = self.session.query(Resource)
        query = query.filter(Resource.id == self.reservations[0].resource)

        return query.one()

    @cached_property
    def reservations(self):
        # libres allows for multiple reservations with a single request (token)
        # for now we don't really have that case in onegov.org, but we
        # try to be aware of it as much as possible
        query = self.session.query(Reservation)
        query = query.filter(Reservation.token == self.id)
        query = query.order_by(Reservation.start)

        return tuple(query)

    @cached_property
    def submission(self):
        return FormSubmissionCollection(self.session).by_id(self.id)

    @property
    def payment(self):
        return self.reservations and self.reservations[0].payment

    @property
    def deleted(self):
        return False if self.reservations else True

    @property
    def extra_data(self):
        return self.submission and [
            v for v in self.submission.data.values()
            if isinstance(v, str)
        ]

    @property
    def email(self):
        # the e-mail is the same over all reservations
        return self.reservations[0].email

    @property
    def undecided(self):
        # if there is no reservation with an 'accept' marker, the user
        # has not yet made a decision
        if self.deleted:
            return False

        for r in self.reservations:
            if r.data and r.data.get('accepted'):
                return False

        return True

    @property
    def title(self):
        parts = []

        for ix, reservation in enumerate(self.reservations):
            parts.append(self.get_reservation_title(reservation))

            if ix == 4:
                parts.append('…')
                break

        return ', '.join(parts)

    def get_reservation_title(self, reservation):
        return self.resource.reservation_title(reservation)

    @property
    def subtitle(self):
        if self.submission:
            return ', '.join(
                p for p in (self.email, self.submission.title) if p)
        elif self.reservations:
            return self.email
        else:
            return None

    @property
    def group(self):
        return self.resource.title

    @classmethod
    def handle_extra_parameters(cls, session, query, extra_parameters):
        if 'allocation_id' in extra_parameters:
            allocations = session.query(Allocation.group)
            allocations = allocations.filter(
                Allocation.id == int(extra_parameters['allocation_id']))

            tokens = session.query(Reservation.token)
            tokens = tokens.filter(
                Reservation.target.in_(allocations.subquery()))

            handler_ids = tuple(t[0].hex for t in tokens)

            if handler_ids:
                query = query.filter(Ticket.handler_id.in_(handler_ids))
            else:
                query = query.filter(False)

        return query

    def get_summary(self, request):
        layout = DefaultLayout(self.resource, request)

        parts = []
        parts.append(
            render_macro(layout.macros['reservations'], request, {
                'reservations': self.reservations,
                'layout': layout
            })
        )

        if self.submission:
            form = self.submission.form_class(data=self.submission.data)

            parts.append(
                render_macro(layout.macros['display_form'], request, {
                    'form': form,
                    'layout': layout
                })
            )

        return ''.join(parts)

    def get_links(self, request):

        if self.deleted:
            return []

        links = []

        accepted = tuple(
            r.data and r.data.get('accepted') or False
            for r in self.reservations
        )

        if not all(accepted):
            links.append(
                Link(
                    text=_("Accept all reservations"),
                    url=request.return_here(
                        request.link(self.reservations[0], 'accept')
                    ),
                    attrs={'class': 'accept-link'}
                )
            )

        advanced_links = []

        if self.submission:
            link = URL(request.link(self.submission))
            link = link.query_param('edit', '')
            link = link.query_param('title', request.translate(
                _("Details about the reservation")))
            link = request.return_here(link.as_string())

            advanced_links.append(
                Link(
                    text=_('Edit details'),
                    url=link,
                    attrs={'class': ('edit-link', 'border')}
                )
            )

        advanced_links.append(Link(
            text=_("Reject all"),
            url=request.return_here(
                request.link(self.reservations[0], 'reject')
            ),
            attrs={'class': 'delete-link'},
            traits=(
                Confirm(
                    _("Do you really want to reject all reservations?"),
                    _("Rejecting these reservations can't be undone."),
                    _("Reject reservations"),
                    _("Cancel")
                ),
                Intercooler(
                    request_method='GET',
                    redirect_after=request.url
                )
            )
        ))

        for reservation in self.reservations:
            link = URL(request.link(reservation, 'reject'))
            link = link.query_param('reservation-id', reservation.id)
            link = request.return_here(link.as_string())

            title = self.get_reservation_title(reservation)
            advanced_links.append(Link(
                text=_("Reject ${title}", mapping={'title': title}),
                url=link,
                attrs={'class': 'delete-link'},
                traits=(
                    Confirm(
                        _("Do you really want to reject this reservation?"),
                        _("Rejecting ${title} can't be undone.", mapping={
                            'title': title
                        }),
                        _("Reject reservation"),
                        _("Cancel")
                    ),
                    Intercooler(
                        request_method='GET',
                        redirect_after=request.url
                    )
                )
            ))

        links.append(LinkGroup(
            _("Advanced"),
            links=advanced_links,
            right_side=False
        ))

        return links


@handlers.registered_handler('EVN')
class EventSubmissionHandler(Handler):

    handler_title = _("Events")

    @cached_property
    def collection(self):
        return EventCollection(self.session)

    @cached_property
    def event(self):
        return self.collection.by_id(self.id)

    @property
    def deleted(self):
        return self.event is None

    @cached_property
    def email(self):
        return self.event.meta.get('submitter_email')

    @property
    def title(self):
        return self.event.title

    @property
    def subtitle(self):
        if self.deleted:
            return None

        parts = (
            self.event.meta.get('submitter_email'),
            '{:%d.%m.%Y %H:%M}'.format(self.event.start)
        )

        return ', '.join(p for p in parts if p)

    @property
    def extra_data(self):
        return [
            self.event.description,
            self.event.title,
            self.event.location
        ]

    @property
    def undecided(self):
        return self.event and self.event.state == 'submitted'

    @cached_property
    def group(self):
        return _("Event")

    def get_summary(self, request):
        layout = EventLayout(self.event, request)
        return render_macro(layout.macros['display_event'], request, {
            'event': self.event,
            'layout': layout
        })

    def get_links(self, request):

        links = []
        layout = EventLayout(self.event, request)

        if self.event and self.event.state == 'submitted':
            links.append(Link(
                text=_("Accept event"),
                url=request.return_here(request.link(self.event, 'publish')),
                attrs={'class': 'accept-link'},
            ))

        if self.event:
            links.append(LinkGroup(_("Advanced"), links=(
                Link(
                    text=_('Edit event'),
                    url=request.return_here(
                        request.link(self.event, 'edit')
                    ),
                    attrs={'class': ('edit-link', 'border')}
                ),
                Link(
                    text=_("Reject event"),
                    url=layout.csrf_protected_url(request.link(self.event)),
                    attrs={'class': ('delete-link')},
                    traits=(
                        Confirm(
                            _("Do you really want to reject this event?"),
                            _("Rejecting this event can't be undone."),
                            _("Reject event"),
                            _("Cancel")
                        ),
                        Intercooler(
                            request_method='DELETE',
                            redirect_after=request.link(self.ticket)
                        )
                    )
                )
            ), right_side=False))

        return links


@handlers.registered_handler('DIR')
class DirectoryEntryHandler(Handler):

    handler_title = _("Directory Entry Submissions")

    @cached_property
    def collection(self):
        return FormSubmissionCollection(self.session)

    @cached_property
    def submission(self):
        return self.collection.by_id(self.id)

    @cached_property
    def form(self):
        return self.submission.form_class(data=self.submission.data)

    @cached_property
    def directory(self):
        if self.submission:
            directory_id = self.submission.meta['directory']
        else:
            directory_id = self.ticket.handler_data['directory']

        return self.session.query(Directory).filter_by(id=directory_id).first()

    @cached_property
    def entry(self):
        if self.submission:
            id = self.submission.meta.get('directory_entry')
        else:
            id = self.ticket.handler_data.get('directory_entry')

        return self.session.query(DirectoryEntry).filter_by(id=id).first()

    @property
    def deleted(self):
        if not self.directory:
            return True

        if self.kind == 'change-request':
            if self.submission:
                data = self.submission.meta
            else:
                data = self.ticket.handler_data

            entry = self.session.query(DirectoryEntry)\
                .filter_by(id=data['directory_entry'])\
                .first()

            if not entry:
                return True

        if self.state == 'adopted':
            name = self.ticket.handler_data.get('entry_name')
            return not self.directory.entry_with_name_exists(name)

        return False

    @property
    def email(self):
        return self.submission.email

    @property
    def title(self):
        return self.submission.title

    @property
    def subtitle(self):
        return None

    @property
    def group(self):
        return self.directory.title

    @property
    def payment(self):
        return self.submission and self.submission.payment

    @property
    def state(self):
        return self.ticket.handler_data.get('state')

    @property
    def extra_data(self):
        return self.submission and [
            v for v in self.submission.data.values()
            if isinstance(v, str)
        ]

    @property
    def undecided(self):
        if not self.directory or self.deleted:
            return False

        return self.state is None

    @property
    def kind(self):
        if self.submission:
            data = self.submission.meta
        else:
            data = self.ticket.handler_data

        if 'change-request' in data.get('extensions', ()):
            return 'change-request'
        else:
            return 'new-entry'

    def get_summary(self, request):
        layout = DefaultLayout(self.submission, request)

        # XXX this is a poor man's request.get_form
        self.form.request = request
        self.form.model = self.submission

        macro = layout.macros['directory_entry_submission']

        return render_macro(macro, request, {
            'form': self.form,
            'layout': layout,
            'handler': self,
        })

    def get_links(self, request):

        links = []

        if not self.directory or self.deleted:
            return links

        if self.state is None:
            links.append(
                Link(
                    text=_("Adopt"),
                    url=request.link(
                        self.directory.submission_action(
                            'adopt', self.submission.id
                        )
                    ),
                    attrs={'class': 'accept-link'},
                    traits=(
                        Intercooler(
                            request_method='POST',
                            redirect_after=request.url
                        ),
                    )
                )
            )

        if self.state == 'adopted':
            links.append(
                Link(
                    text=_("View directory entry"),
                    url=request.class_link(DirectoryEntry, {
                        'directory_name': self.directory.name,
                        'name': self.ticket.handler_data['entry_name']
                    }),
                    attrs={'class': 'view-link'},
                )
            )

        advanced_links = []

        if self.state is None:
            link = URL(request.link(self.submission))
            link = link.query_param('edit', '')
            link = link.query_param('title', request.translate(
                _("Edit details")))
            link = request.return_here(link.as_string())

            advanced_links.append(
                Link(
                    text=_('Edit details'),
                    url=link,
                    attrs={'class': ('edit-link', 'border')}
                )
            )

            advanced_links.append(Link(
                text=_("Reject"),
                url=request.link(
                    self.directory.submission_action(
                        'reject', self.submission.id
                    )
                ),
                attrs={'class': 'delete-link'},
                traits=(
                    Confirm(
                        _("Do you really want to reject this entry?"),
                        _("This cannot be undone."),
                        _("Reject entry"),
                        _("Cancel")
                    ),
                    Intercooler(
                        request_method='POST',
                        redirect_after=request.url
                    )
                )
            ))

        links.append(LinkGroup(
            _("Advanced"),
            links=advanced_links,
            right_side=False
        ))

        return links



""" Contains the base application used by other applications. """

from chameleon import PageTemplate
from collections import defaultdict
from dectate import directive
from more.content_security import SELF
from onegov.core import Framework, utils
from onegov.core.framework import default_content_security_policy
from onegov.core.i18n import default_locale_negotiator
from onegov.core.orm import orm_cached
from onegov.file import DepotApp
from onegov.form import FormApp
from onegov.gis import MapboxApp
from onegov.org import directives
from onegov.org.homepage_widgets import transform_homepage_structure
from onegov.org.initial_content import create_new_organisation
from onegov.org.models import Topic, Organisation, PublicationCollection
from onegov.org.request import OrgRequest
from onegov.org.theme import OrgTheme
from onegov.page import Page, PageCollection
from onegov.pay import PayApp
from onegov.reservation import LibresIntegration
from onegov.search import ElasticsearchApp
from onegov.ticket import TicketCollection
from onegov.user import UserApp
from sqlalchemy import desc


class OrgApp(Framework, LibresIntegration, ElasticsearchApp, MapboxApp,
             DepotApp, PayApp, FormApp, UserApp):

    serve_static_files = True
    request_class = OrgRequest

    #: org directives
    homepage_widget = directive(directives.HomepageWidgetAction)
    export = directive(directives.ExportAction)
    userlinks = directive(directives.UserlinkAction)
    directory_search_widget = directive(directives.DirectorySearchWidgetAction)
    settings_view = directive(directives.SettingsView)

    #: the version of this application (do not change manually!)
    version = '1.6.1'

    def is_allowed_application_id(self, application_id):
        """ Stops onegov.server from ever passing the request to the org
        application, if the schema does not exist. This way we can host
        onegov.org in a way that allows all requests to *.example.org

        If the schema for ``newyork.example.org`` exists, the request is
        handled. If the schema does not exist, the request is not handled.

        Here we basically decide if an org exists or not.

        """
        schema = self.namespace + '-' + application_id

        if schema in self.known_schemas:
            return True

        # block invalid schemas from ever being checked
        if not self.session_manager.is_valid_schema(schema):
            return False

        # if the schema exists, remember it
        if self.session_manager.is_schema_found_on_database(schema):
            self.known_schemas.add(schema)

            return True

        return False

    def configure_application(self, **cfg):
        super().configure_application(**cfg)
        self.known_schemas = set()

        if self.has_database_connection:
            schema_prefix = self.namespace + '-'

            self.known_schemas = set(
                s for s in self.session_manager.list_schemas()
                if s.startswith(schema_prefix)
            )

    def configure_organisation(self, **cfg):
        self.enable_user_registration = cfg.get(
            'enable_user_registration',
            False
        )

        self.enable_yubikey = cfg.get(
            'enable_yubikey',
            False
        )

    @orm_cached(policy='on-table-change:organisations')
    def org(self):
        return self.session().query(Organisation).first()

    @orm_cached(policy='on-table-change:pages')
    def root_pages(self):
        query = PageCollection(self.session()).query(ordered=False)
        query = query.order_by(desc(Page.type), Page.order)
        query = query.filter(Page.parent_id == None)

        return tuple(query)

    @orm_cached(policy='on-table-change:organisations')
    def homepage_template(self):
        structure = self.org.meta.get('homepage_structure')

        if structure:
            return PageTemplate(transform_homepage_structure(self, structure))
        else:
            return PageTemplate('')

    @orm_cached(policy='on-table-change:tickets')
    def ticket_count(self):
        return TicketCollection(self.session()).get_count()

    @orm_cached(policy='on-table-change:pages')
    def homepage_pages(self):
        pages = PageCollection(self.session()).query()
        pages = pages.filter(Topic.type == 'topic')
        pages = pages.filter(Topic.meta['is_visible_on_homepage'] == True)

        result = defaultdict(list)

        for page in pages.all():
            if page.is_visible_on_homepage:
                result[page.root.id].append(page)

        for key in result:
            result[key] = list(sorted(
                result[key],
                key=lambda p: utils.normalize_for_url(p.title)
            ))

        return result

    @orm_cached(policy='on-table-change:files')
    def publications_count(self):
        return PublicationCollection(self.session()).query().count()

    def send_email(self, **kwargs):
        """ Wraps :meth:`onegov.core.framework.Framework.send_email`, setting
        the reply_to address by using the reply address from the organisation
        settings.

        """
        category = kwargs.get('category', 'marketing')

        reply_to = kwargs.pop('reply_to', self.org.meta.get('reply_to', None))
        reply_to = reply_to or self.mail[category]['sender']
        reply_to = "{} <{}>".format(self.org.title, reply_to)

        return super().send_email(reply_to=reply_to, **kwargs)

    @property
    def theme_options(self):
        return self.org.theme_options or {}

    def checkout_button(self, button_label, title, price, email, locale):
        provider = self.default_payment_provider

        if not provider:
            return None

        checkout = {
            'label': button_label,
            'amount': price and price.amount,
            'currency': price and price.currency,
            'email': email,
            'name': self.org.name,
            'description': title,
            'locale': locale.split('_')[0],
            'allowRememberMe': 'false'
        }

        if self.org.square_logo_url:
            checkout['image'] = self.org.square_logo_url

        return provider.checkout_button(**checkout)


@OrgApp.webasset_path()
def get_shared_assets_path():
    return utils.module_path('onegov.shared', 'assets/js')


@OrgApp.setting(section='i18n', name='locales')
def get_i18n_used_locales():
    return {'de_CH', 'fr_CH'}


@OrgApp.setting(section='i18n', name='localedirs')
def get_i18n_localedirs():
    return [
        utils.module_path('onegov.org', 'locale'),
        utils.module_path('onegov.form', 'locale'),
        utils.module_path('onegov.user', 'locale')
    ]


@OrgApp.setting(section='i18n', name='default_locale')
def get_i18n_default_locale():
    return 'de_CH'


@OrgApp.setting(section='i18n', name='locale_negotiator')
def get_locale_negotiator():
    def locale_negotiator(locales, request):
        if request.app.org:
            locales = request.app.org.locales or get_i18n_default_locale()

            if isinstance(locales, str):
                locales = (locales, )

            return default_locale_negotiator(locales, request) or locales[0]
        else:
            return default_locale_negotiator(locales, request)
    return locale_negotiator


@OrgApp.static_directory()
def get_static_directory():
    return 'static'


@OrgApp.template_directory()
def get_template_directory():
    return 'templates'


@OrgApp.setting(section='core', name='theme')
def get_theme():
    return OrgTheme()


@OrgApp.setting(section='content_security_policy', name='default')
def org_content_security_policy():
    policy = default_content_security_policy()

    policy.child_src.add(SELF)
    policy.child_src.add('https://*.youtube.com')
    policy.child_src.add('https://*.vimeo.com')
    policy.child_src.add('https://checkout.stripe.com')

    policy.connect_src.add(SELF)
    policy.connect_src.add('https://checkout.stripe.com')
    policy.connect_src.add('https://sentry.io')
    policy.connect_src.add('https://*.google-analytics.com')
    policy.connect_src.add('https://stats.g.doubleclick.net')
    policy.connect_src.add('https://maps.zg.ch')
    policy.connect_src.add('https://api.mapbox.com')

    return policy


@OrgApp.setting(section='org', name='create_new_organisation')
def get_create_new_organisation_factory():
    return create_new_organisation


@OrgApp.setting(section='org', name='status_mail_roles')
def get_status_mail_roles():
    return ('admin', 'editor')


@OrgApp.setting(section='org', name='ticket_manager_roles')
def get_ticket_manager_roles():
    return ('admin', 'editor')


@OrgApp.setting(section='org', name='require_complete_userprofile')
def get_require_complete_userprofile():
    return False


@OrgApp.setting(section='org', name='is_complete_userprofile')
def get_is_complete_userprofile_handler():
    def is_complete_userprofile(request, username):
        return True

    return is_complete_userprofile


@OrgApp.setting(section='org', name='default_directory_search_widget')
def get_default_directory_search_widget():
    return None


@OrgApp.setting(section='org', name='public_ticket_messages')
def get_public_ticket_messages():
    """ Returns a list of message types which are availble on the ticket
    status page, visible to anyone that knows the unguessable url.

    """

    # do *not* add ticket_note here, those are private!
    return (
        'directory',
        'event',
        'payment',
        'reservation',
        'submission',
        'ticket',
        'ticket_chat',
    )


@OrgApp.webasset_path()
def get_js_path():
    return 'assets/js'


@OrgApp.webasset_path()
def get_css_path():
    return 'assets/css'


@OrgApp.webasset_output()
def get_webasset_output():
    return 'assets/bundles'


@OrgApp.webasset('sortable')
def get_sortable_asset():
    yield 'sortable.js'
    yield 'sortable_custom.js'


@OrgApp.webasset('fullcalendar')
def get_fullcalendar_asset():
    yield 'fullcalendar.css'
    yield 'fullcalendar.js'
    yield 'fullcalendar.de.js'
    yield 'fullcalendar.fr.js'
    yield 'reservationcalendar.jsx'
    yield 'reservationcalendar_custom.js'


@OrgApp.webasset('code_editor')
def get_code_editor_asset():
    yield 'ace.js'
    yield 'ace-mode-form.js'
    yield 'ace-mode-markdown.js'
    yield 'ace-mode-xml.js'
    yield 'ace-mode-yaml.js'
    yield 'ace-theme-tomorrow.js'
    yield 'formcode'
    yield 'code_editor.js'


@OrgApp.webasset('editor')
def get_editor_asset():
    yield 'bufferbuttons.js'
    yield 'definedlinks.js'
    yield 'filemanager.js'
    yield 'imagemanager.js'
    yield 'redactor.de.js'
    yield 'redactor.fr.js'
    yield 'input_with_button.js'
    yield 'editor.js'


@OrgApp.webasset('timeline')
def get_timeline_asset():
    yield 'timeline.jsx'


# do NOT minify the redactor, or the copyright notice goes away, which
# is something we are not allowed to do per our license
@OrgApp.webasset('redactor', filters={'js': None})
def get_redactor_asset():
    yield 'redactor.min.js'
    yield 'redactor.css'


@OrgApp.webasset('upload')
def get_upload_asset():
    yield 'upload.js'


@OrgApp.webasset('editalttext')
def get_editalttext_asset():
    yield 'editalttext.js'


@OrgApp.webasset('prompt')
def get_prompt():
    yield 'prompt.jsx'


@OrgApp.webasset('photoswipe')
def get_photoswipe_asset():
    yield 'photoswipe.css'
    yield 'photoswipe-skin.css'
    yield 'photoswipe.js'
    yield 'photoswipe-ui.js'
    yield 'photoswipe-custom.js'


@OrgApp.webasset('tags-input')
def get_tags_input():
    yield 'tags-input.js'
    yield 'tags-input-setup.js'


@OrgApp.webasset('filedigest')
def get_filehash():
    yield 'asmcrypto-lite.js'
    yield 'filedigest.js'


@OrgApp.webasset('many')
def get_many():
    yield 'many.jsx'


@OrgApp.webasset('monthly-view')
def get_monthly_view():
    yield 'daypicker.js'
    yield 'monthly-view.jsx'


@OrgApp.webasset('common')
def get_common_asset():
    yield 'global.js'
    yield 'polyfills.js'
    yield 'jquery.datetimepicker.css'
    yield 'locale.js'
    yield 'modernizr.js'
    yield 'jquery.js'
    yield 'foundation.js'
    yield 'foundation.alert.js'
    yield 'foundation.dropdown.js'
    yield 'foundation.orbit.js'
    yield 'foundation.reveal.js'
    yield 'foundation.topbar.js'
    yield 'intercooler.js'
    yield 'underscore.js'
    yield 'react.js'
    yield 'react-dom.js'
    yield 'form_dependencies.js'
    yield 'confirm.jsx'
    yield 'typeahead.jsx'
    yield 'pay'
    yield 'moment.js'
    yield 'moment.de-ch.js'
    yield 'moment.fr-ch.js'
    yield 'jquery.datetimepicker.js'
    yield 'jquery.mousewheel.js'
    yield 'jquery.popupoverlay.js'
    yield 'jquery.load.js'
    yield 'videoframe.js'
    yield 'datetimepicker.js'
    yield 'url.js'
    yield 'date-range-selector.js'
    yield 'lazyalttext.js'
    yield 'lazysizes.js'
    yield 'toggle.js'
    yield 'common.js'



from collections import OrderedDict
from onegov.core.security import Public
from onegov.file import File
from onegov.file.utils import name_without_extension
from onegov.org import OrgApp, _
from onegov.org.layout import PublicationLayout
from onegov.org.models import GeneralFileCollection
from onegov.org.models import PublicationCollection
from onegov.core.elements import Link
from sqlalchemy import desc, literal


MONTHS = (
    _("January"),
    _("Feburary"),
    _("March"),
    _("April"),
    _("May"),
    _("June"),
    _("July"),
    _("August"),
    _("September"),
    _("October"),
    _("November"),
    _("December"),
)


@OrgApp.html(model=PublicationCollection, permission=Public,
             template='publications.pt')
def view_publications(self, request):
    layout = PublicationLayout(self, request)

    # filter by year, from latest to first year
    s = self.first_year(layout.timezone) or self.year
    e = layout.today().year
    years = range(e, s - 1, -1)

    filters = {
        'years': tuple(
            Link(
                text=str(year),
                active=year == self.year,
                url=request.link(self.for_year(year=year)),
                rounded=True
            ) for year in years
        )
    }

    # load the publications and bucket them into months
    publications = tuple([] for i in range(12))

    query = self.query().with_entities(
        File.id,
        File.name,
        File.stats,
        File.created.op('AT TIME ZONE')(
            literal(layout.timezone.zone)
        ).label('created'),
    ).order_by(desc(File.created))

    for f in query:
        publications[f.created.month - 1].append(f)

    # group the publications by months, while merging empty months
    today = layout.today()
    grouped = OrderedDict()
    spool = tuple()

    def apply_spool(spool):
        if spool:
            if len(spool) == 1:
                label = spool[0]
            else:
                label = f"{spool[0]} - {spool[-1]}"

            grouped[label] = None

        return tuple()

    for ix, publications in enumerate(reversed(publications)):
        month = 12 - ix

        # exclude current year's months if they are in the future
        if self.year == today.year and month > today.month:
            continue

        if not publications:
            spool = (*spool, request.translate(MONTHS[month - 1]))
            continue

        spool = apply_spool(spool)
        grouped[request.translate(MONTHS[month - 1])] = publications

    apply_spool(spool)

    # link to file and thumbnail by id
    def link(f, name=None):
        return request.class_link(File, {'id': f.id}, name=name)

    return {
        'title': _("Publications"),
        'layout': layout,
        'model': self,
        'filters': filters,
        'link': link,
        'grouped': grouped,
        'name_without_extension': name_without_extension,
        'filedigest_handler': request.class_link(
            GeneralFileCollection, name='digest'
        )
    }



import csv
import os.path
import subprocess
import tempfile

from onegov.core.csv import CSVFile


class DigirezDB(object):
    """ Offers access to a Digirez Room Booking Software Database
    (see http://www.digiappz.com).

    Expects that mdbtools are installed (http://mdbtools.sourceforge.net).

    """

    def __init__(self, accessdb_path):
        self.accessdb_path = accessdb_path
        self.csv_directory = None

    @property
    def tables(self):
        output = subprocess.check_output(('mdb-tables', self.accessdb_path))
        output = output.decode('utf-8').rstrip('\n ')

        return output.split(' ')

    @property
    def opened(self):
        return self.csv_directory is not None

    @property
    def csv_path(self):
        return self.csv_directory and self.csv_directory.name

    def open(self):
        assert not self.opened
        self.csv_directory = tempfile.TemporaryDirectory()

        for table in self.tables:
            output_path = os.path.join(self.csv_path, '{}.csv'.format(table))

            with open(output_path, 'w') as output_file:
                subprocess.check_call(
                    args=(
                        'mdb-export', '-D', '%Y-%m-%dT%T',
                        self.accessdb_path, table
                    ),
                    stdout=output_file)

    @property
    def records(self):
        assert self.opened
        return RecordsAccessor(self.csv_path)


class RecordsAccessor(object):

    def __init__(self, csv_path):
        self.csv_path = csv_path
        self.files = {}

    def get_file(self, name):
        if name not in self.files:
            path = os.path.join(self.csv_path, '{}.csv'.format(name))
            self.files[name] = open(path, 'rb')

        return self.files[name]

    def __getattr__(self, name):
        csv_file = CSVFile(
            self.get_file(name),
            expected_headers=None,
            dialect=csv.unix_dialect
        )

        return csv_file.lines



""" Provides commands used to initialize org websites. """

import click
import html
import isodate
import re
import requests
import shutil
import textwrap

from cached_property import cached_property
from collections import defaultdict
from datetime import date, datetime, timedelta
from libres.db.models import ReservedSlot
from libres.modules.errors import InvalidEmailAddress, AlreadyReservedError
from onegov.core.custom import json
from onegov.core.cache import lru_cache
from onegov.core.cli import command_group, pass_group_context, abort
from onegov.core.csv import CSVFile
from onegov.form import FormCollection
from onegov.org.formats import DigirezDB
from onegov.org.models import Organisation, TicketNote
from onegov.org.forms import ReservationForm
from onegov.reservation import Allocation, Reservation
from onegov.reservation import ResourceCollection
from onegov.ticket import TicketCollection
from onegov.form import as_internal_id, parse_form
from onegov.user import UserCollection, User
from purl import URL
from sedate import replace_timezone, utcnow
from sqlalchemy import create_engine, event, text
from sqlalchemy.orm import Session
from tqdm import tqdm
from uuid import UUID, uuid4


cli = command_group()


@cli.command(context_settings={'creates_path': True})
@click.argument('name')
@click.option('--locale',
              default='de_CH', type=click.Choice(['de_CH', 'fr_CH']))
@pass_group_context
def add(group_context, name, locale):
    """ Adds an org with the given name to the database. For example:

        onegov-org --select '/onegov_org/evilcorp' add "Evilcorp"

    """

    def add_org(request, app):

        if app.session().query(Organisation).first():
            abort("{} already contains an organisation".format(
                group_context.selector))

        app.settings.org.create_new_organisation(app, name, locale=locale)

        click.echo("{} was created successfully".format(name))

    return add_org


@cli.command()
@pass_group_context
def delete(group_context):
    """ Deletes a single organisation matching the selector.

    Selectors matching multiple organisations are disabled for saftey reasons.

    """

    def delete_org(request, app):

        org = app.org.title
        confirmation = "Do you really want to DELETE {}?".format(org)

        if not click.confirm(confirmation):
            abort("Deletion process aborted")

        if app.has_filestorage:
            click.echo("Removing File Storage")

            for item in app.filestorage.listdir('.'):
                if app.filestorage.isdir(item):
                    app.filestorage.removedir(item, recursive=True, force=True)
                else:
                    app.filestorage.remove(item)

        if app.depot_storage_path and app.bound_storage_path:
            click.echo("Removing Depot Storage")
            shutil.rmtree(str(app.bound_storage_path.absolute()))

        if app.has_database_connection:
            click.echo("Dropping Database Schema")

            assert app.session_manager.is_valid_schema(app.schema)

            dsn = app.session_manager.dsn
            app.session_manager.session().close_all()
            app.session_manager.dispose()

            engine = create_engine(dsn)
            engine.execute('DROP SCHEMA "{}" CASCADE'.format(app.schema))
            engine.raw_connection().invalidate()
            engine.dispose()

        click.echo("{} was deleted successfully".format(org))

    return delete_org


@cli.command(name='import-digirez', context_settings={'singular': True})
@click.argument('accessdb', type=click.Path(exists=True, resolve_path=True))
@click.option('--min-date', default=None,
              help="Min date in the form '2016-12-31'")
@click.option('--ignore-booking-conflicts', default=False, is_flag=True,
              help="Ignore booking conflicts (TESTING ONlY)")
def import_digirez(accessdb, min_date, ignore_booking_conflicts):
    """ Imports a Digirez reservation database into onegov.org.

    Example:

        onegov-org --select '/orgs/govikon' import-digirez room_booking.mdb

    """

    if not shutil.which('mdb-export'):
        abort("Could not find 'mdb-export', please install mdbtools!")

    min_date = min_date and isodate.parse_date(min_date) or date.today()
    min_date = datetime(min_date.year, min_date.month, min_date.day)

    db = DigirezDB(accessdb)
    db.open()

    # XXX this is currently specific to what Gemeinde Rüti has, we could put
    # this into an external file however.
    form_definition = textwrap.dedent("""
        Name *= ___
        Telefon =___
        Zweck *= ___
        Einrichtungen =
            [ ] Konzertbestuhlung
            [ ] Bankettbestuhlung
            [ ] Flipchart
            [ ] Hellraumprojektor
            [ ] Kinoleinwand (Saal)
            [ ] Mobile Leinwand
            [ ] Lautsprecheranlage
            [ ] Podestelemente
            [ ] Flügel (Saal)
            [ ] Getränke
            [ ] Küche (Saal)
            [ ] Office (Keller)
        Bemerkungen = ...
    """)

    floor_hours = {
        'Amthaus': (8, 24),
        'Löwen': (0, 24),
        'Gemeinschaftszentrum': (8, 24)
    }

    title_map = {
        'Jugend- raum': 'Jugendraum',
        'Spielgruppen- Raum': 'Spielgruppenraum',
        'Vereins- raum': 'Vereinsraum'
    }

    def get_formdata(member, booking):

        einrichtungen = []

        if booking.kbest == '1':
            einrichtungen.append('Konzertbestuhlung')

        if booking.bbest == '1':
            einrichtungen.append('Bankettbestuhlung')

        if booking.flip == '1':
            einrichtungen.append('Flipchart')

        if booking.ohproj == '1':
            einrichtungen.append('Hellraumprojektor')

        if booking.klw == '1':
            einrichtungen.append('Kinoleinwand (Saal)')

        if booking.mlw == '1':
            einrichtungen.append('Mobile Leinwand')

        if booking.lspr == '1':
            einrichtungen.append('Lautsprecheranlage')

        if booking.pod == '1':
            einrichtungen.append('Podestelemente')

        if booking.flg == '1':
            einrichtungen.append('Flügel (Saal)')

        if booking.getr == '1':
            einrichtungen.append('Getränke')

        if booking.getr == '1':
            einrichtungen.append('Küche (Saal)')

        if booking.ofk == '1':
            einrichtungen.append('Office (Keller)')

        return {
            'name': booking.anspname or ' '.join((
                member.member_last_name,
                member.member_first_name
            )),
            'telefon': booking.ansptel or member.member_phone,
            'zweck': booking.title,
            'bemerkungen': booking.description,
            'einrichtungen': einrichtungen
        }

    def unescape_dictionary(d):
        for key, value in d.items():
            if isinstance(value, str):
                d[key] = html.unescape(value)
            if isinstance(value, dict):
                d[key] = unescape_dictionary(value)
            if isinstance(value, list):
                d[key] = [html.unescape(v) for v in value]

        return d

    email_fixes = (
        (re.compile(r'2da-capo-rueti.ch$'), '@da-capo.rueti.ch'),
        (re.compile(r'@mzol$'), '@mzol.ch'),
        (re.compile(r'^jo$'), 'info@rueti.ch'),
        (re.compile(r'@pro-vitalis$'), '@pro-vitalis.ch')
    )

    def run_import(request, app):

        # create all resources first, fails if at least one exists already
        print("Creating resources")

        resources = ResourceCollection(app.libres_context)
        floors = {f.id: f.floor_name for f in db.records.floors}
        resources_by_room = {}

        for room in db.records.room:
            resource = resources.add(
                title=title_map.get(room.room_name, room.room_name),
                timezone='Europe/Zurich',
                type='room',
                group=floors[room.floor_id],
                definition=form_definition,
                meta={'is_hidden_from_public': True}
            )

            resources_by_room[room.id] = resource

        # gather all information needed to create the allocations/reservations
        relevant_bookings = (
            b for b in db.records.room_booking
            if isodate.parse_datetime(b.hour_end) > min_date
        )

        # group by room id in addition to the multi_id, as digirez supports
        # reservations which span multiple rooms
        reservations = defaultdict(list)

        for booking in relevant_bookings:
            group = '-'.join((booking.room_id, booking.multi_id))
            reservations[group].append(booking)

        # get the max_date per resource
        max_dates = {}

        for reservation_group in reservations:
            resource_id = resources_by_room[reservation_group.split('-')[0]].id

            if resource_id not in max_dates:
                max_dates[resource_id] = min_date

            for booking in reservations[reservation_group]:
                date = isodate.parse_datetime(booking.hour_end)

                if max_dates[resource_id] < date:
                    max_dates[resource_id] = date

        # get the user associated with the migration (the first admin)
        users = UserCollection(app.session())
        user = users.query().filter(User.role == 'admin').first()

        if not user:
            abort('No admin user found')

        # get the member mails
        members = {}

        for member in db.records.members:
            members[member.member_id] = member

        # create an allocation for all days between min/max date
        print("Creating allocations")

        for resource in resources_by_room.values():

            # true if the resource does not have any bookings
            if resource.id not in max_dates:
                continue

            start_hour, end_hour = floor_hours[resource.group]
            days = (max_dates[resource.id].date() - min_date.date()).days

            first_day_start = datetime(
                min_date.year, min_date.month, min_date.day,
                start_hour
            )

            first_day_end = datetime(
                min_date.year, min_date.month, min_date.day,
                end_hour - 1, 59, 59, 999999
            )

            whole_day = (start_hour, end_hour) == (0, 24)

            for day in range(0, days + 1):
                start = first_day_start + timedelta(days=day)
                end = first_day_end + timedelta(days=day)

                resource.scheduler.allocate(
                    dates=(start, end),
                    partly_available=True,
                    whole_day=whole_day,
                )

        # create the reservations
        print("Creating reservations")

        booking_conflicts = 0

        for reservation_group, bookings in reservations.items():
            resource = resources_by_room[reservation_group.split('-')[0]]
            scheduler = resource.scheduler
            found_conflict = False

            session_id = uuid4()

            for booking in bookings:
                assert bookings[0].anspmail == booking.anspmail
                assert bookings[0].member_id == booking.member_id
                assert bookings[0].anspname == booking.anspname
                assert bookings[0].ansptel == booking.ansptel
                assert bookings[0].title == booking.title
                assert bookings[0].description == booking.description

                email = booking.anspmail
                email = email or members[booking.member_id].member_email

                for expression, replacement in email_fixes:
                    email = expression.sub(replacement, email)

                try:
                    token = scheduler.reserve(
                        email=email,
                        dates=(
                            isodate.parse_datetime(booking.hour_start),
                            isodate.parse_datetime(booking.hour_end)
                        ),
                        session_id=session_id,
                        single_token_per_session=True,
                        data={'accepted': True}  # accepted through ticket
                    )
                    token = token.hex
                except InvalidEmailAddress:
                    abort("{} is an invalid e-mail address".format(email))
                except AlreadyReservedError:
                    booking_conflicts += 1
                    found_conflict = True

                    print("Booking conflict in {} at {}".format(
                        resource.title, booking.hour_start
                    ))
                    pass

            if found_conflict:
                continue

            forms = FormCollection(app.session())
            form_data = get_formdata(members[booking.member_id], booking)
            form_data = unescape_dictionary(form_data)
            form = resource.form_class(data=form_data)

            if not form.validate():
                abort("{} failed the form check with {}".format(
                    form_data, form.errors
                ))

            submission = forms.submissions.add_external(
                form=form,
                state='pending',
                id=token
            )

            scheduler.queries.confirm_reservations_for_session(session_id)
            scheduler.approve_reservations(token)

            forms.submissions.complete_submission(submission)

            with forms.session.no_autoflush:
                ticket = TicketCollection(request.session).open_ticket(
                    handler_code='RSV', handler_id=token
                )
                ticket.accept_ticket(user)
                ticket.close_ticket()

        if not ignore_booking_conflicts and booking_conflicts:
            abort("There were {} booking conflicts, aborting".format(
                booking_conflicts
            ))

    return run_import


@cli.command(name='import-reservations', context_settings={'singular': True})
@click.option('--dsn', required=True, help="DSN to the source database")
@click.option('--map', required=True, help="CSV map between resources")
def import_reservations(dsn, map):
    """ Imports reservations from a legacy seantis.reservation system.

    WARNING: Existing reservations and all associated tickets/submissions
    are deleted during the migration.

    Pending reservations are ignored. The expectation is that the old system
    is stopped, then migrated, then never started again.

    Also note that search reindexing is disabled during import. A manual
    reindex is necessary afterwards.

    Example:

    onegov-org --select '/orgs/govikon' import-reservations\
        --dsn postgresql://localhost:5432/legacy\
        --map ressource-map.csv

    The ressource map is a simple CSV file with the following columns:

    * Legacy URL (the URL pointing to the legacy ressource/room)
    * OGC URL (the URL pointing to the associated ressource in the OGC)
    * Type ('daypass' or 'room')

    The first row is expected to be the header row.

    """

    print(f"Connecting to remote")
    engine = create_engine(dsn).connect()

    @event.listens_for(engine, 'begin')
    def receive_begin(conn):
        conn.execute('SET TRANSACTION READ ONLY')

    remote = Session(bind=engine)

    # define the mapping between old/new resources
    class Mapping(object):

        def __init__(self, libres_context, old_url, new_url, type):
            self.resources = ResourceCollection(libres_context)
            self.old_url = old_url
            self.new_url = new_url
            self.type = type

            if type != 'room':
                raise NotImplementedError(
                    "Room migration has not been tested yet")

        @property
        def name(self):
            return URL(self.new_url).path().rstrip('/').split('/')[-1]

        @cached_property
        def resource(self):
            return self.resources.by_name(name=self.name)

        @cached_property
        def old_uuid(self):
            return UUID(requests.get(f'{self.old_url}/@@uuid').text.strip())

        @property
        def new_uuid(self):
            return self.resource.id

    # run the given query returning the query and a count (if possible)
    def select_with_count(session, query, **params):
        query = query.strip(' \n')
        assert query.startswith("SELECT *")

        count = text(
            query
            .replace("SELECT *", "SELECT COUNT(*)", 1)
            .split('ORDER BY')[0]
        )
        count = session.execute(count, params).scalar()

        return count, session.execute(text(query), params)

    print(f"Reading map")
    records = CSVFile(open(map, 'rb'), ('Old URL', 'New URL', 'Type')).lines
    records = tuple(records)

    # generate forms from the form-data found in the external system
    def get_form_class_and_data(data):

        if not data:
            return ReservationForm, {}

        def exclude(value):
            return as_internal_id(value['desc']) == 'email'

        @lru_cache()
        def generate_form_class(formcode):
            return parse_form(formcode, ReservationForm)

        def separate_code_from_data(data):
            fields = []
            values = {}

            for form in sorted(data.values(), key=lambda f: f['desc']):
                fieldset = form['desc']
                fields.append(f"# {fieldset}")

                for v in sorted(form['values'], key=lambda r: r['sortkey']):
                    label = v['desc']
                    id = as_internal_id(f"{fieldset} {label}")

                    # defined on the reservation form
                    if id == 'email':
                        continue

                    if isinstance(v['value'], bool):
                        fields.append(f"{label} = ___")
                        values[id] = v['value'] and "Ja" or "Nein"
                    elif isinstance(v['value'], str):
                        fields.append(f"{label} = ___")
                        values[id] = v['value']
                    elif isinstance(v['value'], datetime):
                        fields.append(f"{label} = YYYY.MM.DD HH:MM")
                        values[id] = v['value']
                    elif isinstance(v['value'], date):
                        fields.append(f"{label} = YYYY.MM.DD")
                        values[id] = v['value']
                    elif isinstance(v['value'], int):
                        fields.append(f"{label} = ___")
                        values[id] = str(v['value'])
                    else:
                        raise NotImplementedError((
                            f"No conversion for {v['value']} "
                            f" ({type(v['value'])}"
                        ))

            return '\n'.join(fields), values

        formcode, formdata = separate_code_from_data(data)
        return generate_form_class(formcode), formdata

    def handle_import(request, app):
        session = app.session()

        # disable search indexing during import
        if hasattr(app, 'es_orm_events'):
            app.es_orm_events.stopped = True

        # map the old UUIDs to the resources
        print("Mapping resources")
        mapping = {m.old_uuid: m for m in (
            Mapping(request.app.libres_context, r.old_url, r.new_url, r.type)
            for r in tqdm(records, unit=' resources')
        )}

        print("Clearing existing submissions")
        session.execute(text("""
            DELETE FROM submissions
             WHERE submissions.meta->>'origin' IN :resources
        """), {
            'resources': tuple(m.old_uuid.hex for m in mapping.values())
        })

        print("Clearing existing ticket messages")
        session.execute(text("""
            DELETE from messages
             WHERE channel_id IN (
                SELECT number
                  FROM tickets
                 WHERE tickets.handler_data->>'origin' IN :resources
             )
        """), {
            'resources': tuple(m.old_uuid.hex for m in mapping.values())
        })

        print("Clearing existing tickets")
        session.execute(text("""
            DELETE FROM tickets
             WHERE tickets.handler_data->>'origin' IN :resources
        """), {
            'resources': tuple(m.old_uuid.hex for m in mapping.values())
        })

        payment_ids = tuple(r[0] for r in session.execute(text("""
            SELECT payment_id
            FROM payments_for_reservations_payment
            WHERE reservations_id IN (
                SELECT id
                FROM reservations
                WHERE resource IN :resources
            )
        """), {
            'resources': tuple(m.resource.id for m in mapping.values())
        }))

        if payment_ids:
            print("Clearing existing payments")
            session.execute(text("""
                DELETE FROM payments_for_reservations_payment
                WHERE payment_id IN :payments
            """), {
                'payments': payment_ids
            })

            session.execute(text("""
                DELETE FROM payments
                WHERE id IN :payments
            """), {
                'payments': payment_ids
            })

        print("Clearing existing reservations")
        session.execute(text("""
            DELETE FROM reservations
            WHERE resource IN :resources
        """), {
            'resources': tuple(m.resource.id for m in mapping.values())
        })

        print("Clearing existing reserved slots")
        session.execute(text("""
            DELETE FROM reserved_slots
             WHERE allocation_id IN (
                SELECT id FROM allocations
                 WHERE allocations.mirror_of IN :resources
             )
        """), {
            'resources': tuple(m.resource.id for m in mapping.values())
        })

        print("Clearing existing allocations")
        session.execute(text("""
            DELETE FROM allocations
             WHERE allocations.mirror_of IN :resources
        """), {
            'resources': tuple(m.resource.id for m in mapping.values())
        })

        session.flush()

        print("Fetching remote allocations")
        count, rows = select_with_count(remote, """
            SELECT * FROM allocations
            WHERE mirror_of IN :resources
            ORDER BY allocations.resource
        """, resources=tuple(mapping.keys()))

        # we use a separate id space, so we need to keep track
        allocation_ids = {}

        # the resource might be mapped, but it is not a given
        def row_resource(row):
            if row['resource'] not in mapping:
                return row['resource']

            return mapping[row['resource']].new_uuid

        # create the new allocations
        print("Writing allocations")
        for row in tqdm(rows, unit=' allocations', total=count):
            resource_type = mapping[row['mirror_of']].type

            if row['partly_available'] and resource_type != 'room':
                raise RuntimeError((
                    f"Cannot migrate partly_available allocation "
                    f"to a {resource_type} resource"
                ))

            if row['approve_manually']:
                raise RuntimeError((
                    f"Manually approved allocation found (id: {row['id']}), "
                    f"manually approved allocations are not supported"
                ))

            allocation = Allocation(
                resource=row_resource(row),
                mirror_of=mapping[row['mirror_of']].new_uuid,
                group=row['group'],
                quota=row['quota'],
                quota_limit=row['quota_limit'],
                partly_available=row['partly_available'],
                approve_manually=row['approve_manually'],

                # the timezone was ignored in seantis.reservation
                timezone='Europe/Zurich',
                _start=replace_timezone(row['_start'], 'Europe/Zurich'),
                _end=replace_timezone(row['_end'], 'Europe/Zurich'),

                data=json.loads(row['data']),
                _raster=row['_raster'],
                created=replace_timezone(row['created'], 'UTC'),
                modified=utcnow(),
                type='custom',
            )

            session.add(allocation)
            session.flush()

            allocation_ids[row['id']] = allocation.id

        # fetch the reserved slots that should be migrated
        count, rows = select_with_count(remote, """
            SELECT * FROM reserved_slots WHERE allocation_id IN (
                SELECT id FROM allocations WHERE mirror_of IN :resources
            )
        """, resources=tuple(mapping.keys()))

        # create the reserved slots with the mapped values
        print("Writing reserved slots")
        for row in tqdm(rows, unit=" slots", total=count):
            session.add(ReservedSlot(
                resource=mapping[row['resource']].new_uuid,
                start=replace_timezone(row['start'], 'Europe/Zurich'),
                end=replace_timezone(row['end'], 'Europe/Zurich'),
                allocation_id=allocation_ids[row['allocation_id']],
                reservation_token=row['reservation_token']
            ))

        session.flush()

        # fetch the reservations that should be migrated
        count, rows = select_with_count(remote, """
            SELECT * FROM reservations
            WHERE resource IN :resources
              AND status = 'approved'
            ORDER BY resource
        """, resources=tuple(mapping.keys()))

        def targeted_allocations(group):
            return session.query(Allocation).filter_by(group=group)

        # keep track of custom reservation data, for the creation of tickets
        reservation_data = {}

        print("Writing reservations")
        for row in tqdm(rows, unit=' reservations', total=count):

            reservation_data[row['token']] = {
                'data': json.loads(row['data']),
                'email': row['email'],
                'origin': row['resource'],
                'origin_url': mapping[row['resource']].old_url,
                'created': replace_timezone(row['created'], 'UTC'),
                'modified': replace_timezone(row['modified'], 'UTC'),
            }

            if row['quota'] > 1:
                raise NotImplementedError((
                    "Reservations with a quota > 1 are not supported yet"
                ))

            # onegov.reservation does not support group targets, so we
            # translate those into normal allocations and create multiple
            # reservations with a shared token
            shared = dict(
                token=row['token'],
                target_type='allocation',
                resource=mapping[row['resource']].new_uuid,
                timezone='Europe/Zurich',
                status=row['status'],
                data={"accepted": True, "migrated": True},
                email=row['email'],
                quota=row['quota'],
                created=replace_timezone(row['created'], 'UTC'),
                modified=replace_timezone(row['modified'], 'UTC'),
                type='custom',
            )

            if row['target_type'] == 'group':
                for allocation in targeted_allocations(group=row['target']):
                    allocation.group = uuid4()

                    session.add(Reservation(
                        target=allocation.group,
                        start=allocation.start,
                        end=allocation.end,
                        **shared
                    ))

            else:
                session.add(Reservation(
                    target=row['target'],
                    start=replace_timezone(row['start'], 'Europe/Zurich'),
                    end=replace_timezone(row['end'], 'Europe/Zurich'),
                    **shared
                ))

        session.flush()

        # tie reservations to tickets/submissions
        tickets = TicketCollection(session)
        forms = FormCollection(session)

        # the responsible user is the first admin that was added
        user = session.query(User)\
            .filter_by(role='admin')\
            .order_by(User.created).first()

        print("Writing tickets")
        for token, data in tqdm(reservation_data.items(), unit=" tickets"):
            form_class, form_data = get_form_class_and_data(data['data'])

            if form_data:
                form_data['email'] = data['email']
                form = form_class(data=form_data)

                # wtforms requires raw_data for some validators
                for key, value in form_data.items():
                    getattr(form, key).raw_data = [value]

                submission = forms.submissions.add_external(
                    form=form,
                    state='complete',
                    id=token
                )

                submission.meta['migrated'] = True
                submission.meta['origin'] = data['origin'].hex

            with session.no_autoflush:
                ticket = tickets.open_ticket(
                    handler_code='RSV', handler_id=token.hex)

                ticket.handler_data['migrated'] = True
                ticket.handler_data['origin'] = data['origin'].hex
                ticket.handler_data['origin_url'] = data['origin_url']
                ticket.muted = True
                ticket.state = 'closed'
                ticket.last_state_change = ticket.timestamp()
                ticket.reaction_time = 0
                ticket.user = user
                ticket.created = data['created']
                ticket.modified = data['modified']

                TicketNote.create(ticket, request, (
                    f"Migriert von {data['origin_url']}"
                    f"/reservations?token={token}",
                ), owner=user.username)

    return handle_import



""" Contains the model describing the page editor. """


class Editor(object):
    """ Defines the model for the page editor. Required because pages need
    to be edited outside their url structure, since their urls are absorbed
    completely and turned into SQL queries.

    """
    def __init__(self, action, page, trait=None):
        """ The editor is defined by an action and a page/context.

        :action:
            One of 'new', 'edit' or 'delete'.

        :page:
            The 'context' of the action. The actual page in the case of 'edit'
            and 'delete'. The parent in the case of 'new' or 'paste'.

            New pages inherit the type from the parent.

        :trait:
            The trait of the page. Currently either 'link' or 'page'.
            Only necessary if it's a new page. The trait controls the content
            of the page and leads to different forms.

            See :mod:`onegov.org.models.page`.

        """

        assert self.is_supported_action(action)

        self.action = action
        self.page = page
        self.trait = action == 'new' and trait or page.trait

    @staticmethod
    def is_supported_action(action):
        """ Returns True if the given action is supported. """
        return action in {'new', 'paste', 'edit', 'delete'}

    @property
    def page_id(self):
        """ Returns the page id so morepath can create a link to this. """
        return self.page.id



import re
import transaction

from collections import namedtuple
from onegov.core.security import Public, Private, Secret
from onegov.core.utils import render_file
from onegov.directory import Directory
from onegov.directory import DirectoryCollection
from onegov.directory import DirectoryEntry
from onegov.directory import DirectoryEntryCollection
from onegov.directory import DirectoryZipArchive
from onegov.directory.errors import DuplicateEntryError
from onegov.directory.errors import MissingColumnError
from onegov.directory.errors import MissingFileError
from onegov.directory.errors import ValidationError
from onegov.file import File
from onegov.form import FormCollection, as_internal_id
from onegov.form.fields import UploadField
from onegov.org import OrgApp, _
from onegov.org.forms import DirectoryForm, DirectoryImportForm
from onegov.org.forms.generic import ExportForm
from onegov.org.layout import DirectoryCollectionLayout
from onegov.org.layout import DirectoryEntryCollectionLayout
from onegov.org.layout import DirectoryEntryLayout
from onegov.org.models import DirectorySubmissionAction
from onegov.org.models import ExtendedDirectory, ExtendedDirectoryEntry
from onegov.core.elements import Link
from purl import URL
from tempfile import NamedTemporaryFile
from webob.exc import HTTPForbidden


def get_directory_form_class(model, request):
    return ExtendedDirectory().with_content_extensions(DirectoryForm, request)


def get_directory_entry_form_class(model, request):
    form_class = ExtendedDirectoryEntry().with_content_extensions(
        model.directory.form_class, request)

    class OptionalMapForm(form_class):
        def on_request(self):
            if model.directory.enable_map == 'no':
                self.delete_field('coordinates')

    return OptionalMapForm


def get_submission_form_class(model, request):
    return model.directory.form_class_for_submissions(include_private=True)


def get_change_request_form_class(model, request):
    return model.directory.form_class_for_submissions(include_private=False)


@OrgApp.html(
    model=DirectoryCollection,
    template='directories.pt',
    permission=Public)
def view_directories(self, request):
    return {
        'title': _("Directories"),
        'layout': DirectoryCollectionLayout(self, request),
        'directories': request.exclude_invisible(self.query()),
        'link': lambda directory: request.link(
            DirectoryEntryCollection(directory)
        )
    }


@OrgApp.view(
    model=Directory,
    permission=Public)
def view_directory_redirect(self, request):
    return request.redirect(request.class_link(
        DirectoryEntryCollection, {'directory_name': self.name}
    ))


@OrgApp.form(model=DirectoryCollection, name='new', template='form.pt',
             permission=Secret, form=get_directory_form_class)
def handle_new_directory(self, request, form):
    if form.submitted(request):
        directory = self.add_by_form(form, properties=('configuration', ))

        request.success(_("Added a new directory"))
        return request.redirect(
            request.link(DirectoryEntryCollection(directory)))

    layout = DirectoryCollectionLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Directories"), request.link(self)),
        Link(_("New"), request.link(self, name='new'))
    ]

    return {
        'layout': layout,
        'title': _("New Directory"),
        'form': form,
        'form_width': 'huge',
    }


@OrgApp.form(model=DirectoryEntryCollection, name='edit',
             template='directory_form.pt', permission=Secret,
             form=get_directory_form_class)
def handle_edit_directory(self, request, form):
    migration = None
    error = None

    if form.submitted(request):
        save_changes = True

        if self.directory.entries:
            migration = self.directory.migration(
                form.structure.data,
                form.configuration
            )

            if migration.changes:
                if not migration.possible:
                    save_changes = False
                    request.alert(_(
                        "The requested change cannot be performed, "
                        "as it is incompatible with existing entries"
                    ))
                else:
                    if not request.params.get('confirm'):
                        form.action += '&confirm=1'
                        save_changes = False

        if save_changes:
            form.populate_obj(self.directory)

            try:
                self.session.flush()
            except ValidationError as e:
                error = e
                error.link = request.class_link(DirectoryEntry, {
                    'directory_name': self.directory.name,
                    'name': e.entry.name
                })
                transaction.abort()
            else:
                request.success(_("Your changes were saved"))
                return request.redirect(request.link(self))

    elif not request.POST:
        form.process(obj=self.directory)

    layout = DirectoryCollectionLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Directories"), request.link(self)),
        Link(_(self.directory.title), request.link(self)),
        Link(_("Edit"), '#')
    ]

    return {
        'layout': layout,
        'title': self.directory.title,
        'form': form,
        'form_width': 'large',
        'migration': migration,
        'model': self,
        'error': error,
        'error_translate': lambda text: request.translate(_(text)),
        'directory': self.directory,
    }


@OrgApp.view(
    model=DirectoryEntryCollection,
    permission=Secret,
    request_method='DELETE')
def delete_directory(self, request):
    request.assert_valid_csrf_token()

    session = request.session

    for entry in self.directory.entries:
        session.delete(entry)

    DirectoryCollection(session).delete(self.directory)
    request.success(_("The directory was deleted"))


@OrgApp.html(
    model=DirectoryEntryCollection,
    permission=Public,
    template='directory.pt')
def view_directory(self, request):

    Filter = namedtuple('Filter', ('title', 'tags'))
    filters = []
    empty = tuple()

    radio_fields = set(
        f.id for f in self.directory.fields if f.type == 'radio'
    )

    for keyword, title, values in self.available_filters:
        filters.append(Filter(title=title, tags=tuple(
            Link(
                text=value,
                active=value in self.keywords.get(keyword, empty),
                url=request.link(self.for_filter(
                    singular=keyword in radio_fields,
                    **{keyword: value}
                )),
                rounded=keyword in radio_fields
            ) for value in values
        )))

    entries = request.exclude_invisible(self.query())

    if self.directory.configuration.thumbnail:
        thumbnail = as_internal_id(self.directory.configuration.thumbnail)
    else:
        thumbnail = None

    def thumbnail_link(entry):
        id = (entry.values.get(thumbnail) or {}).get('data', '').lstrip('@')
        return id and request.class_link(File, {'id': id}, name='thumbnail')

    return {
        'layout': DirectoryEntryCollectionLayout(self, request),
        'title': self.directory.title,
        'entries': entries,
        'directory': self.directory,
        'searchwidget': self.searchwidget,
        'filters': filters,
        'geojson': request.link(self, name='+geojson'),
        'submit': request.link(self, name='+submit'),
        'show_thumbnails': thumbnail and True or False,
        'thumbnail_link': thumbnail_link
    }


@OrgApp.json(
    model=DirectoryEntryCollection,
    permission=Public,
    name='geojson')
def view_geojson(self, request):
    q = self.query()
    q = q.with_entities(
        DirectoryEntry.id,
        DirectoryEntry.name,
        DirectoryEntry.title,
        DirectoryEntry.lead,
        DirectoryEntry.content["coordinates"]["lat"].label('lat'),
        DirectoryEntry.content["coordinates"]["lon"].label('lon'),
        DirectoryEntry.meta["is_hidden_from_public"].label('is_hidden')
    )
    q = q.filter(DirectoryEntry.content["coordinates"]["lat"] != None)

    # this could be done using a query, but that seems to be more verbose
    entries = (c for c in q if request.is_manager or not c.is_hidden)

    url_prefix = request.class_link(DirectoryEntry, {
        'directory_name': self.directory.name,
        'name': ''
    })

    return tuple({
        'type': "Feature",
        'properties': {
            'name': e.name,
            'title': e.title,
            'lead': e.lead,
            'link': url_prefix + e.name
        },
        'geometry': {
            'coordinates': [e.lon, e.lat],
            'type': "Point"
        }
    } for e in entries)


@OrgApp.form(
    model=DirectoryEntryCollection,
    permission=Private,
    template='form.pt',
    form=get_directory_entry_form_class,
    name='new')
def handle_new_directory_entry(self, request, form):
    if form.submitted(request):
        entry = self.directory.add_by_form(form, type='extended')

        request.success(_("Added a new directory entry"))
        return request.redirect(request.link(entry))

    if form.errors:
        for field in form.match_fields(include_classes=(UploadField, )):
            getattr(form, field).data = {}

    layout = DirectoryEntryCollectionLayout(self, request)
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(_("New"), '#'))
    layout.editbar_links = []

    return {
        'layout': layout,
        'title': _("New Directory Entry"),
        'form': form,
    }


@OrgApp.form(
    model=DirectoryEntry,
    permission=Private,
    template='form.pt',
    form=get_directory_entry_form_class,
    name='edit')
def handle_edit_directory_entry(self, request, form):
    if form.submitted(request):
        form.populate_obj(self)

        request.success(_("Your changes were saved"))
        return request.redirect(request.link(self))
    elif not request.POST:
        form.process(obj=self)

    layout = DirectoryEntryLayout(self, request)
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(_("Edit"), '#'))
    layout.editbar_links = []

    return {
        'layout': layout,
        'title': self.title,
        'form': form,
    }


@OrgApp.form(model=DirectoryEntryCollection,
             permission=Public,
             template='directory_entry_submission_form.pt',
             form=get_submission_form_class,
             name='submit')
def handle_submit_directory_entry(self, request, form):

    if not self.directory.enable_submissions:
        raise HTTPForbidden()

    title = _("Submit a New Directory Entry")

    if form.submitted(request):
        forms = FormCollection(request.session)

        # required by the form submissions collection
        form._source = self.directory.structure

        # the price per submission
        if self.directory.price == 'paid':
            amount = self.directory.price_per_submission
        else:
            amount = 0.0

        submission = forms.submissions.add_external(
            form=form,
            state='pending',
            payment_method=self.directory.payment_method,
            email=form.submitter.data,
            meta={
                'handler_code': 'DIR',
                'directory': self.directory.id.hex,
                'price': {
                    'amount': amount,
                    'currency': self.directory.currency
                },
                'extensions': tuple(
                    ext for ext in self.directory.extensions
                    if ext != 'submitter'
                )
            }
        )

        # remove old submission while we are at it
        self.directory.remove_old_pending_submissions()

        url = URL(request.link(submission))
        url = url.query_param('title', request.translate(title))

        return request.redirect(url.as_string())

    layout = DirectoryEntryCollectionLayout(self, request)
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(title, '#'))
    layout.editbar_links = []

    return {
        'directory': self.directory,
        'form': form,
        'layout': layout,
        'title': title,
        'guideline': self.directory.submissions_guideline,
    }


@OrgApp.form(model=DirectoryEntry,
             permission=Public,
             template='directory_entry_submission_form.pt',
             form=get_change_request_form_class,
             name='change-request')
def handle_change_request(self, request, form):

    if not self.directory.enable_change_requests:
        raise HTTPForbidden()

    title = _("Propose a change")

    if form.submitted(request):
        forms = FormCollection(request.session)

        # required by the form submissions collection
        form._source = self.directory.structure

        extensions = [
            ext for ext in self.directory.extensions if ext != 'submitter']
        extensions.append('change-request')

        submission = forms.submissions.add_external(
            form=form,
            state='pending',
            email=form.submitter.data,
            meta={
                'handler_code': 'DIR',
                'directory': self.directory.id.hex,
                'directory_entry': self.id.hex,
                'extensions': extensions,
            }
        )

        # remove old submission while we are at it
        self.directory.remove_old_pending_submissions()

        url = URL(request.link(submission))
        url = url.query_param('title', request.translate(title))

        return request.redirect(url.as_string())

    elif not request.POST:
        form.process(obj=self)

    layout = DirectoryEntryLayout(self, request)
    layout.include_code_editor()
    layout.breadcrumbs.append(Link(title, '#'))
    layout.editbar_links = []

    return {
        'directory': self.directory,
        'form': form,
        'layout': layout,
        'title': title,
        'hint': _(
            "To request a change, edit the fields you would like to change, "
            "leaving the other fields intact. Then submit your request."
        ),
        'guideline': self.directory.change_requests_guideline,
    }


@OrgApp.html(
    model=DirectoryEntry,
    permission=Public,
    template='directory_entry.pt')
def view_directory_entry(self, request):

    return {
        'layout': DirectoryEntryLayout(self, request),
        'title': self.title,
        'entry': self
    }


@OrgApp.view(
    model=DirectoryEntry,
    permission=Private,
    request_method='DELETE')
def delete_directory_entry(self, request):
    request.assert_valid_csrf_token()

    session = request.session
    session.delete(self)

    request.success(_("The entry was deleted"))


@OrgApp.form(model=DirectoryEntryCollection, permission=Private, name='export',
             template='export.pt', form=ExportForm)
def view_export(self, request, form):

    layout = DirectoryEntryCollectionLayout(self, request)
    layout.breadcrumbs.append(Link(_("Export"), '#'))
    layout.editbar_links = None

    if form.submitted(request):
        url = URL(request.link(self, '+zip'))
        url = url.query_param('format', form.format)

        return request.redirect(url.as_string())

    return {
        'layout': layout,
        'title': _("Export"),
        'form': form,
        'explanation': _(
            "Exports all entries of this directory. The resulting zipfile "
            "contains the selected format as well as metadata and "
            "images/files if the directory contains any."
        )
    }


@OrgApp.view(model=DirectoryEntryCollection, permission=Private, name='zip')
def view_zip_file(self, request):
    layout = DirectoryEntryCollectionLayout(self, request)

    format = request.params.get('format', 'json')
    formatter = layout.export_formatter(format)

    def transform(key, value):
        return formatter(key), formatter(value)

    with NamedTemporaryFile() as f:
        archive = DirectoryZipArchive(f.name + '.zip', format, transform)
        archive.write(self.directory)

        response = render_file(str(archive.path), request)

    filename = ' '.join((
        self.directory.name, layout.format_date(layout.now(), 'datetime')))
    filename = re.sub(r'[\.:]+', '-', filename)

    response.headers['Content-Disposition']\
        = 'attachment; filename="{}"'.format(filename)

    return response


@OrgApp.form(model=DirectoryEntryCollection, permission=Private, name='import',
             template='directory_import.pt', form=DirectoryImportForm)
def view_import(self, request, form):
    error = None

    layout = DirectoryEntryCollectionLayout(self, request)
    layout.breadcrumbs.append(Link(_("Import"), '#'))
    layout.editbar_links = None

    if form.submitted(request):
        try:
            imported = form.run_import(target=self.directory)
        except MissingColumnError as e:
            request.alert(_("The column ${name} is missing", mapping={
                'name': self.directory.field_by_id(e.column).human_id
            }))
        except MissingFileError as e:
            request.alert(_("The file ${name} is missing", mapping={
                'name': e.name
            }))
        except DuplicateEntryError as e:
            request.alert(_("The entry ${name} exists twice", mapping={
                'name': e.name
            }))
        except ValidationError as e:
            error = e
        except NotImplementedError:
            request.alert(_(
                "The given file is invalid, does it include a metadata.json "
                "with a data.xlsx, data.csv, or data.json?"
            ))
        else:
            notify = imported and request.success or request.warning
            notify(_("Imported ${count} entries", mapping={
                'count': imported
            }))

            return request.redirect(request.link(self))

        # no success if we land here
        transaction.abort()

    return {
        'layout': layout,
        'title': _("Import"),
        'form': form,
        'explanation': _(
            "Updates the directory configuration and imports all entries "
            "given in the ZIP file. The format is the same as produced by "
            "the export function. Note that only 100 items are imported at a "
            "time. To import more items repeat the import accordingly."
        ),
        'directory': self.directory,
        'error': error,
        'error_translate': lambda text: request.translate(_(text)),
    }


@OrgApp.view(
    model=DirectorySubmissionAction,
    permission=Private,
    request_method='POST'
)
def execute_submission_action(self, request):
    return self.execute(request)



""" Contains the models describing files and images. """

import sedate

from cached_property import cached_property
from collections import namedtuple
from datetime import datetime
from dateutil.relativedelta import relativedelta
from itertools import chain, groupby
from onegov.core.orm import as_selectable
from onegov.core.orm.mixins import meta_property
from onegov.file import File, FileSet, FileCollection, FileSetCollection
from onegov.file import SearchableFile
from onegov.file.utils import IMAGE_MIME_TYPES_AND_SVG
from onegov.org import _
from onegov.org.models.extensions import HiddenFromPublicExtension
from onegov.search import ORMSearchable
from sedate import standardize_date, utcnow
from sqlalchemy import asc, desc, select


DateInterval = namedtuple('DateInterval', ('name', 'start', 'end'))


class GroupFilesByDateMixin(object):

    def get_date_intervals(self, today):
        today = standardize_date(today, 'UTC')

        month_end = today + relativedelta(day=31)
        month_start = today - relativedelta(day=1)

        yield DateInterval(
            name=_("This month"),
            start=month_start,
            end=month_end)

        last_month_end = month_start - relativedelta(microseconds=1)
        last_month_start = month_start - relativedelta(months=1)

        yield DateInterval(
            name=_("Last month"),
            start=last_month_start,
            end=last_month_end)

        if month_start.month not in (1, 2):
            this_year_end = last_month_start - relativedelta(microseconds=1)
            this_year_start = this_year_end.replace(
                month=1, day=1, hour=0, minute=0, second=0, microsecond=0)

            yield DateInterval(
                name=_("This year"),
                start=this_year_start,
                end=this_year_end)
        else:
            this_year_end = None
            this_year_start = None

        last_year_end = this_year_start or last_month_start
        last_year_end -= relativedelta(microseconds=1)
        last_year_start = last_year_end.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0)

        yield DateInterval(
            name=_("Last year"),
            start=last_year_start,
            end=last_year_end)

        older_end = last_year_start - relativedelta(microseconds=1)
        older_start = datetime(2000, 1, 1, tzinfo=today.tzinfo)

        yield DateInterval(
            name=_("Older"),
            start=older_start,
            end=older_end)

    def query_intervals(self, intervals, before_filter=None, process=None):
        base_query = self.query().order_by(desc(File.created))

        if before_filter:
            base_query = before_filter(base_query)

        for interval in intervals:
            query = base_query.filter(File.created >= interval.start)
            query = query.filter(File.created <= interval.end)

            for result in query.all():
                if process is not None:
                    yield interval.name, process(result)

    def grouped_by_date(self, today=None, id_only=True):
        """ Returns all files grouped by natural language dates.

        By default, only ids are returned, as this is enough to build the
        necessary links, which is what you usually want from a file.

        The given date is expected to be in UTC.

        """

        intervals = tuple(self.get_date_intervals(today or utcnow()))

        def before_filter(query):
            if id_only:
                return query.with_entities(File.id)

            return query

        def process(result):
            if id_only:
                return result.id

            return result

        return groupby(
            self.query_intervals(intervals, before_filter, process),
            key=lambda item: item[0]
        )


class GeneralFile(File, SearchableFile):
    __mapper_args__ = {'polymorphic_identity': 'general'}

    @property
    def es_public(self):
        return self.published and self.signed


class ImageFile(File):
    __mapper_args__ = {'polymorphic_identity': 'image'}


class ImageSet(FileSet, HiddenFromPublicExtension, ORMSearchable):
    __mapper_args__ = {'polymorphic_identity': 'image'}

    es_properties = {
        'title': {'type': 'localized'},
        'lead': {'type': 'localized'}
    }

    @property
    def es_public(self):
        return not self.is_hidden_from_public

    @property
    def es_suggestions(self):
        return {
            "input": [self.title.lower()]
        }

    lead = meta_property()
    view = meta_property()

    show_images_on_homepage = meta_property()


class ImageSetCollection(FileSetCollection):

    def __init__(self, session):
        super().__init__(session, type='image')


class GeneralFileCollection(FileCollection, GroupFilesByDateMixin):

    supported_content_types = 'all'

    file_list = as_selectable("""
        SELECT
            row_number() OVER () as number, -- Integer
            id,                             -- Text
            name,                           -- Text
            "order",                        -- Text
            signed,                         -- Boolean
            created as upload_date,         -- UTCDateTime
            reference->>'content_type'
                AS content_type             -- Text
        FROM files
        WHERE type = 'general'
    """)

    def __init__(self, session, order_by='name'):
        super().__init__(session, type='general', allow_duplicates=False)

        self.order_by = order_by
        self.direction = order_by == 'name' and 'ascending' or 'descending'

        self._last_interval = None

    def for_order(self, order):
        return self.__class__(self.session, order_by=order)

    @cached_property
    def intervals(self):
        return tuple(self.get_date_intervals(today=sedate.utcnow()))

    @property
    def statement(self):
        stmt = select(self.file_list.c)

        if self.order_by == 'name':
            order = self.file_list.c.order
        else:
            order = self.file_list.c.upload_date

        direction = self.direction == 'ascending' and asc or desc

        return stmt.order_by(direction(order))

    @property
    def files(self):
        return self.session.execute(self.statement)

    def group(self, record):
        if self.order_by == 'name':
            if record.order[0].isdigit():
                return '0-9'

            return record.order[0].upper()
        else:
            if self._last_interval:
                intervals = chain((self._last_interval, ), self.intervals)
            else:
                intervals = self.intervals

            for interval in intervals:
                if interval.start <= record.upload_date <= interval.end:
                    break
            else:
                return _("Older")

            # this method is usually called for each item in a sorted set,
            # we optimise for that by caching the last matching interval
            # and checking that one first the next time
            self._last_interval = interval

            return interval.name


class ImageFileCollection(FileCollection, GroupFilesByDateMixin):

    supported_content_types = IMAGE_MIME_TYPES_AND_SVG

    def __init__(self, session):
        super().__init__(session, type='image', allow_duplicates=False)



""" The authentication views. """

import morepath

from onegov.core.markdown import render_untrusted_markdown
from onegov.core.security import Public, Personal
from onegov.org import _, OrgApp
from onegov.org import log
from onegov.org.elements import Link
from onegov.org.layout import DefaultLayout
from onegov.org.mail import send_transactional_html_mail
from onegov.user import Auth, UserCollection
from onegov.user.errors import AlreadyActivatedError
from onegov.user.errors import ExistingUserError
from onegov.user.errors import ExpiredSignupLinkError
from onegov.user.errors import InvalidActivationTokenError
from onegov.user.errors import UnknownUserError
from onegov.user.forms import LoginForm
from onegov.user.forms import PasswordResetForm
from onegov.user.forms import RegistrationForm
from onegov.user.forms import RequestPasswordResetForm
from purl import URL
from webob import exc


@OrgApp.form(model=Auth, name='login', template='login.pt', permission=Public,
             form=LoginForm)
def handle_login(self, request, form):
    """ Handles the login requests. """

    if not request.app.enable_yubikey:
        form.delete_field('yubikey')

    if self.skippable(request):
        return self.redirect(request)

    if form.submitted(request):

        redirected_to_userprofile = False

        org_settings = request.app.settings.org
        if org_settings.require_complete_userprofile:
            username = form.username.data

            if not org_settings.is_complete_userprofile(request, username):
                redirected_to_userprofile = True

                self.to = request.return_to(
                    '/userprofile',
                    self.to
                )

        response = self.login_to(request=request, **form.login_data)

        if response:
            if redirected_to_userprofile:
                request.warning(_(
                    "Your userprofile is incomplete. "
                    "Please update it before you continue."
                ))
            else:
                request.success(_("You have been logged in."))

            return response

        request.alert(_("Wrong e-mail address, password or yubikey."))

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Login"), request.link(self, name='login'))
    ]

    def provider_login(provider):
        provider.to = self.to
        return request.link(provider)

    return {
        'layout': layout,
        'password_reset_link': request.link(self, name='request-password'),
        'register_link': request.link(self, name='register'),
        'may_register': request.app.enable_user_registration,
        'button_text': _("Login"),
        'providers': request.app.providers,
        'provider_login': provider_login,
        'render_untrusted_markdown': render_untrusted_markdown,
        'title': _('Login to ${org}', mapping={
            'org': request.app.org.title
        }),
        'form': form
    }


@OrgApp.form(model=Auth, name='register', template='form.pt',
             permission=Public, form=RegistrationForm)
def handle_registration(self, request, form):
    """ Handles the user registration. """

    if not request.app.enable_user_registration:
        raise exc.HTTPNotFound()

    if form.submitted(request):

        try:
            user = self.register(form, request)
        except ExistingUserError:
            request.alert(_("A user with this address already exists"))
        except ExpiredSignupLinkError:
            request.alert(_("This signup link has expired"))
        else:
            url = URL(request.link(self, 'activate'))
            url = url.query_param('username', form.username.data)
            url = url.query_param('token', user.data['activation_token'])

            subject = request.translate(
                _("Your ${org} Registration", mapping={
                    'org': request.app.org.title
                })
            )

            send_transactional_html_mail(
                request=request,
                template='mail_activation.pt',
                subject=subject,
                receivers=(form.username.data, ),
                content={
                    'activation_link': url.as_string(),
                    'model': self
                }
            )
            request.success(_(
                "Thank you for registering. Please follow the instructions "
                "on the activiation e-mail sent to you."
            ))

            return morepath.redirect(request.link(request.app.org))

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Register"), request.link(self, name='register'))
    ]

    return {
        'layout': layout,
        'title': _('Account Registration'),
        'form': form
    }


@OrgApp.view(model=Auth, name='activate', permission=Public)
def handle_activation(self, request):

    if not request.app.enable_user_registration:
        raise exc.HTTPNotFound()

    users = UserCollection(request.session)

    username = request.params.get('username')
    token = request.params.get('token')

    try:
        users.activate_with_token(username, token)
    except UnknownUserError:
        request.warning(_("Unknown user"))
    except InvalidActivationTokenError:
        request.warning(_("Invalid activation token"))
    except AlreadyActivatedError:
        request.success(_("Your account has already been activated."))
    else:
        request.success(_(
            "Your account has been activated. "
            "You may now log in with your credentials"
        ))

    return morepath.redirect(request.link(request.app.org))


@OrgApp.html(model=Auth, name='logout', permission=Personal)
def view_logout(self, request):
    """ Handles the logout requests. """

    return self.logout_to(request)


@OrgApp.form(model=Auth, name='request-password', template='form.pt',
             permission=Public, form=RequestPasswordResetForm)
def handle_password_reset_request(self, request, form):
    """ Handles the GET and POST password reset requests. """

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Reset password"), request.link(self, name='request-password'))
    ]

    if form.submitted(request):

        user = UserCollection(request.session)\
            .by_username(form.email.data)

        url = layout.password_reset_url(user)

        if url:
            send_transactional_html_mail(
                request=request,
                template='mail_password_reset.pt',
                subject=_("Password reset"),
                receivers=(user.username, ),
                content={'model': None, 'url': url}
            )
        else:
            log.info(
                "Failed password reset attempt by {}".format(
                    request.client_addr
                )
            )

        response = morepath.redirect(request.link(self, name='login'))
        request.success(
            _(('A password reset link has been sent to ${email}, provided an '
               'account exists for this email address.'),
              mapping={'email': form.email.data})
        )
        return response

    return {
        'layout': layout,
        'title': _('Reset password'),
        'form': form,
        'form_width': 'small'
    }


@OrgApp.form(model=Auth, name='reset-password', template='form.pt',
             permission=Public, form=PasswordResetForm)
def handle_password_reset(self, request, form):
    if form.submitted(request):
        # do NOT log the user in at this point - only onegov.user.auth does
        # logins - we only ever want one path to be able to login, which makes
        # it easier to do it correctly.

        if form.update_password(request):
            request.success(_("Password changed."))
            return morepath.redirect(request.link(self, name='login'))
        else:
            request.alert(
                _("Wrong username or password reset link not valid any more.")
            )
            log.info(
                "Failed password reset attempt by {}".format(
                    request.client_addr
                )
            )

    if 'token' in request.params:
        form.token.data = request.params['token']

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Reset password"), request.link(self, name='request-password'))
    ]

    return {
        'layout': layout,
        'title': _('Reset password'),
        'form': form,
        'form_width': 'small'
    }



from collections import OrderedDict
from functools import partial
from onegov.core.security import Private
from onegov.form import merge_forms
from onegov.org import OrgApp, _
from onegov.org.forms import DateRangeForm, ExportForm
from onegov.org.layout import PaymentCollectionLayout
from onegov.org.mail import send_ticket_mail
from onegov.org.models import PaymentMessage
from onegov.core.elements import Link
from sedate import align_range_to_day, standardize_date, as_datetime
from onegov.pay import Payment
from onegov.pay import PaymentCollection
from onegov.pay import PaymentProviderCollection
from onegov.ticket import TicketCollection


EMAIL_SUBJECTS = {
    'marked-as-paid': _("Your payment has been received"),
    'marked-as-unpaid': _("Your payment has been withdrawn"),
    'refunded': _("Your payment has been refunded")
}


def ticket_by_link(tickets, link):
    if link.__tablename__ == 'reservations':
        return tickets.by_handler_id(link.token.hex)
    elif link.__tablename__ == 'submissions':
        return tickets.by_handler_id(link.id.hex)


def send_ticket_notifications(payment, request, change):
    session = request.session
    tickets = TicketCollection(session)

    for link in payment.links:
        ticket = ticket_by_link(tickets, link)

        if not ticket:
            continue

        # create a notification in the chat
        PaymentMessage.create(payment, ticket, request, change)

        if change == 'captured':
            continue

        # send an e-mail
        email = ticket.snapshot.get('email') or ticket.handler.email

        send_ticket_mail(
            request=request,
            template='mail_payment_change.pt',
            subject=EMAIL_SUBJECTS[change],
            receivers=(email, ),
            ticket=ticket,
            content={
                'model': ticket,
                'payment': payment,
                'change': change
            }
        )


@OrgApp.html(
    model=PaymentCollection,
    template='payments.pt',
    permission=Private)
def view_payments(self, request):
    tickets = TicketCollection(request.session)

    providers = {
        provider.id: provider
        for provider in PaymentProviderCollection(request.session).query()
    }

    payment_links = self.payment_links_by_batch()

    return {
        'title': _("Payments"),
        'layout': PaymentCollectionLayout(self, request),
        'payments': self.batch,
        'get_ticket': partial(ticket_by_link, tickets),
        'providers': providers,
        'payment_links': payment_links
    }


@OrgApp.form(
    model=PaymentCollection,
    name='export',
    template='form.pt',
    permission=Private,
    form=merge_forms(DateRangeForm, ExportForm))
def export_payments(self, request, form):
    layout = PaymentCollectionLayout(self, request)
    layout.breadcrumbs.append(Link(_("Export"), '#'))
    layout.editbar_links = None

    if form.submitted(request):
        start, end = align_range_to_day(
            standardize_date(as_datetime(form.data['start']), layout.timezone),
            standardize_date(as_datetime(form.data['end']), layout.timezone),
            layout.timezone)

        return form.as_export_response(
            run_export(
                session=request.session,
                start=start,
                end=end,
                nested=form.format == 'json',
                formatter=layout.export_formatter(form.format)
            )
        )

    return {
        'title': _("Export"),
        'layout': layout,
        'form': form
    }


def run_export(session, start, end, nested, formatter):
    collection = PaymentCollection(session, start=start, end=end)

    payments = tuple(collection.subset())
    links = collection.payment_links_by_batch(payments)

    def transform(payment, links):
        r = OrderedDict()
        r['source'] = payment.source
        r['source_id'] = payment.remote_id
        r['state'] = payment.state
        r['currency'] = payment.currency
        r['gross'] = payment.amount
        r['net'] = round(payment.net_amount, 2)
        r['fee'] = round(payment.fee, 2)
        r['payout_id'] = payment.meta.get('payout_id')
        r['payout_date'] = payment.meta.get('payout_date')

        if nested:
            r['references'] = [l.payable_reference for l in links]
        else:
            r['references'] = '\n'.join(l.payable_reference for l in links)

        for key in r:
            r[key] = formatter(r[key])

        return r

    return tuple(transform(p, links[p.id]) for p in payments)


@OrgApp.view(
    model=Payment,
    name='mark-as-paid',
    request_method='POST',
    permission=Private)
def mark_as_paid(self, request):
    request.assert_valid_csrf_token()
    send_ticket_notifications(self, request, 'marked-as-paid')

    request.success(_("The ticket was marked as paid"))

    assert self.source == 'manual'
    self.state = 'paid'


@OrgApp.view(
    model=Payment,
    name='mark-as-unpaid',
    request_method='POST',
    permission=Private)
def mark_as_unpaid(self, request):
    request.assert_valid_csrf_token()
    send_ticket_notifications(self, request, 'marked-as-unpaid')

    request.success(_("The ticket was marked as unpaid"))

    assert self.source == 'manual'
    self.state = 'open'


@OrgApp.view(
    model=Payment,
    name='capture',
    request_method='POST',
    permission=Private)
def capture(self, request):
    request.assert_valid_csrf_token()
    send_ticket_notifications(self, request, 'captured')

    request.success(_("The payment was captured"))

    assert self.source == 'stripe_connect'
    self.charge.capture()


@OrgApp.view(
    model=Payment,
    name='refund',
    request_method='POST',
    permission=Private)
def refund(self, request):
    request.assert_valid_csrf_token()
    send_ticket_notifications(self, request, 'refunded')

    request.success(_("The payment was refunded"))

    assert self.source == 'stripe_connect'
    self.refund()



""" Newsletter subscription management. """

import morepath

from onegov.core.security import Public
from onegov.newsletter import NewsletterCollection, Subscription
from onegov.org import _, OrgApp


# use an english name for this view, so robots know what we use it for
@OrgApp.view(model=Subscription, name='confirm', permission=Public)
def view_confirm(self, request):
    if self.confirm():
        request.success(_(
            "the subscription for ${address} was successfully confirmed",
            mapping={'address': self.recipient.address}
        ))
    else:
        request.alert(_(
            "the subscription for ${address} could not be confirmed, "
            "wrong token",
            mapping={'address': self.recipient.address}
        ))

    return morepath.redirect(
        request.link(NewsletterCollection(request.session))
    )


# use an english name for this view, so robots know what we use it for
@OrgApp.view(model=Subscription, name='unsubscribe', permission=Public)
def view_unsubscribe(self, request):

    address = self.recipient.address

    if self.unsubscribe():
        request.success(_(
            "${address} successfully unsubscribed",
            mapping={'address': address}
        ))
    else:
        request.alert(_(
            "${address} could not be unsubscribed, wrong token",
            mapping={'address': address}
        ))

    return morepath.redirect(
        request.link(NewsletterCollection(request.session))
    )



from onegov.core.templates import render_template
from onegov.org.layout import DefaultMailLayout


def send_html_mail(request, template, content, **kwargs):
    """" Sends an email rendered from the given template.

    Example::

        send_html_mail(request, 'mail_template.pt', {'model': self},
            subject=_("Test subject")
            receivers=('recipient@example.org', )
        )

    """

    assert 'model' in content
    assert 'subject' in kwargs
    assert 'receivers' in kwargs

    kwargs['subject'] = request.translate(kwargs['subject'])

    if 'layout' not in content:
        content['layout'] = DefaultMailLayout(content['model'], request)

    if 'title' not in content:
        content['title'] = kwargs['subject']

    kwargs['content'] = render_template(template, request, content)

    # the email is queued here, not actually sent!
    request.app.send_email(**kwargs)


def send_transactional_html_mail(*args, **kwargs):
    kwargs['category'] = 'transactional'
    return send_html_mail(*args, **kwargs)


def send_marketing_html_mail(*args, **kwargs):
    kwargs['category'] = 'marketing'
    return send_html_mail(*args, **kwargs)


def send_ticket_mail(request, template, subject, receivers, ticket,
                     content=None, force=False, **kwargs):

    if not force:

        if ticket.muted:
            return

        if request.current_username in receivers:
            if len(receivers) == 1:
                return

            receivers = tuple(
                r for r in receivers if r != request.current_username
            )

    subject = ticket.reference(request) + ': ' + request.translate(subject)

    content = content or {}

    # legacy behavior
    if 'model' not in content:
        content['model'] = ticket

    if 'ticket' not in content:
        content['ticket'] = ticket

    return send_transactional_html_mail(
        request=request,
        template=template,
        subject=subject,
        receivers=receivers,
        content=content,
        **kwargs)



from cached_property import cached_property
from onegov.core.html_diff import render_html_diff
from onegov.form.extensions import FormExtension
from onegov.form.fields import UploadField
from onegov.gis import CoordinatesField
from onegov.org import _
from wtforms.fields import TextAreaField
from wtforms.fields.html5 import EmailField
from wtforms.validators import DataRequired


class CoordinatesFormExtension(FormExtension, name='coordinates'):

    def create(self):
        class CoordinatesForm(self.form_class):
            coordinates = CoordinatesField(
                label=_("Coordinates"),
                description=_(
                    "The marker can be moved by dragging it with the mouse"
                ),
                fieldset=_("Map"),
                render_kw={'data-map-type': 'marker'}
            )

        return CoordinatesForm


class SubmitterFormExtension(FormExtension, name='submitter'):

    def create(self):
        class SubmitterForm(self.form_class):
            submitter = EmailField(
                label=_("E-Mail"),
                fieldset=_("Submitter"),
                validators=[DataRequired()]
            )

        return SubmitterForm


class CommentFormExtension(FormExtension, name='comment'):

    def create(self):
        class CommentForm(self.form_class):
            comment = TextAreaField(
                label=_("Comment"),
                fieldset=_("Submitter"),
                render_kw={'rows': 7}
            )

        return CommentForm


class ChangeRequestFormExtension(FormExtension, name='change-request'):

    def create(self):

        # XXX circular import
        from onegov.org.models.directory import ExtendedDirectoryEntry

        class ChangeRequestForm(self.form_class):

            @cached_property
            def target(self):

                # not all steps have this information set, for example, towards
                # the end, the onegov.form submission code runs an extra
                # validation, which we ignore, trusting that it all worked
                # out earlier
                if not getattr(self, 'model', None):
                    return None

                return self.request.session.query(ExtendedDirectoryEntry)\
                    .filter_by(id=self.model.meta['directory_entry'])\
                    .first()

            def is_different(self, field):
                # if the target has been removed, stop
                if not self.target:
                    return True

                # after the changes have been applied, use the list of changes
                if self.model.meta.get('changed'):
                    return field.id in self.model.meta['changed']

                # ignore CSRF token
                if field.id == 'csrf_token':
                    return False

                # coordinates fields are provided through extension
                if field.id == 'coordinates':
                    return field.data != self.target.coordinates

                # upload fields differ if they are not empty
                if isinstance(field, UploadField):
                    return field.data and True or False

                return self.target.values.get(field.id) != field.data

            def render_original(self, field):
                prev = field.data

                try:
                    field.data = self.target.values.get(field.id)
                    return super().render_display(field)
                finally:
                    field.data = prev

            def render_display(self, field):
                if self.is_different(field):
                    proposed = super().render_display(field)

                    if not self.target:
                        return proposed

                    if field.id in ('csrf_token', 'coordinates'):
                        return proposed

                    if field.id not in self.target.values:
                        return proposed

                    if isinstance(field, UploadField):
                        return proposed

                    original = self.render_original(field)
                    return render_html_diff(original, proposed)

            def ensure_changes(self):
                if not self.target:
                    return

                for name, field in self._fields.items():
                    if self.is_different(field):
                        return

                self.errors['global-errors'] = [
                    _("Please provide at least one change")]

                return False

        return ChangeRequestForm



import sedate

from datetime import datetime
from onegov.core.orm.mixins import meta_property, content_property
from onegov.core.orm.types import UUID
from onegov.form.models import FormSubmission
from onegov.org.models.extensions import ContactExtension
from onegov.org.models.extensions import CoordinatesExtension
from onegov.org.models.extensions import HiddenFromPublicExtension
from onegov.org.models.extensions import PersonLinkExtension
from onegov.reservation import Resource, Reservation
from onegov.search import SearchableContent
from onegov.ticket import Ticket
from sqlalchemy.orm import undefer
from sqlalchemy.sql.expression import cast
from uuid import uuid4, uuid5


class SharedMethods(object):

    lead = meta_property()
    text = content_property()

    @property
    def deletable(self):
        if self.scheduler.managed_reserved_slots().first():
            return False

        if self.scheduler.managed_reservations().first():
            return False

        return True

    @property
    def calendar_date_range(self):
        """ Returns the date range set by the fullcalendar specific params. """

        if self.date:
            date = datetime(self.date.year, self.date.month, self.date.day)
            date = sedate.replace_timezone(date, self.timezone)
        else:
            date = sedate.to_timezone(sedate.utcnow(), self.timezone)

        if self.view == 'month':
            return sedate.align_range_to_month(date, date, self.timezone)
        elif self.view == 'agendaWeek':
            return sedate.align_range_to_week(date, date, self.timezone)
        elif self.view == 'agendaDay':
            return sedate.align_range_to_day(date, date, self.timezone)

    def remove_expired_reservation_sessions(self, expiration_date=None):
        session = self.libres_context.get_service('session_provider').session()
        queries = self.scheduler.queries

        expired_sessions = queries.find_expired_reservation_sessions(
            expiration_date)

        if expired_sessions:
            query = session.query(Reservation).with_entities(Reservation.token)
            query = query.filter(Reservation.session_id.in_(expired_sessions))
            tokens = set(r[0] for r in query.all())

            query = session.query(FormSubmission)
            query = query.filter(FormSubmission.name == None)
            query = query.filter(FormSubmission.id.in_(tokens))

            query.delete('fetch')
            queries.remove_expired_reservation_sessions(expiration_date)

    def bound_reservations(self, request, status='pending'):
        """ The reservations associated with this resource and user. """

        session = self.bound_session_id(request)
        scheduler = self.get_scheduler(request.app.libres_context)

        res = scheduler.queries.reservations_by_session(session)
        res = res.filter(Reservation.resource == self.id)
        res = res.filter(Reservation.status == status)
        res = res.order_by(False)  # clear existing order
        res = res.order_by(Reservation.start)

        # used by ReservationInfo
        res = res.options(undefer(Reservation.created))

        return res

    def bound_session_id(self, request):
        """ The session id associated with this resource and user. """
        if not request.browser_session.has('libres_session_id'):
            request.browser_session.libres_session_id = uuid4()

        return uuid5(self.id, request.browser_session.libres_session_id.hex)

    def reservations_with_tickets_query(self, start, end):
        """ Returns a query which joins this resource's reservations between
        start and end with the tickets table.

        """
        query = self.scheduler.managed_reservations()
        query = query.filter(start <= Reservation.start)
        query = query.filter(Reservation.end <= end)

        query = query.join(
            Ticket, Reservation.token == cast(Ticket.handler_id, UUID))

        query = query.order_by(Reservation.start)
        query = query.order_by(Ticket.subtitle)
        query = query.filter(Reservation.status == 'approved')
        query = query.filter(Reservation.data != None)

        return query

    def reservation_title(self, reservation):
        title = self.title_template.format(
            start=reservation.display_start(),
            end=reservation.display_end(),
            quota=reservation.quota
        )

        if title.endswith('00:00'):
            return title[:-5] + '24:00'

        return title


class DaypassResource(Resource, HiddenFromPublicExtension, SearchableContent,
                      ContactExtension, PersonLinkExtension,
                      CoordinatesExtension, SharedMethods):
    __mapper_args__ = {'polymorphic_identity': 'daypass'}

    es_type_name = 'daypasses'

    # the default view
    view = 'month'

    # show or hide quota numbers in reports
    show_quota = True

    # use to render the reservation title
    title_template = '{start:%d.%m.%Y} ({quota})'


class RoomResource(Resource, HiddenFromPublicExtension, SearchableContent,
                   ContactExtension, PersonLinkExtension,
                   CoordinatesExtension, SharedMethods):
    __mapper_args__ = {'polymorphic_identity': 'room'}

    es_type_name = 'rooms'

    # the default view
    view = 'agendaWeek'

    # show or hide quota numbers in reports
    show_quota = False

    # used to render the reservation title
    title_template = '{start:%d.%m.%Y} {start:%H:%M} - {end:%H:%M}'



import json

from cached_property import cached_property
from datetime import date, datetime, timedelta
from dateutil import rrule
from dateutil.rrule import rrulestr
from onegov.event.models import EventFile
from onegov.form import Form
from onegov.form.fields import MultiCheckboxField
from onegov.form.fields import UploadFileWithORMSupport
from onegov.form.validators import FileSizeLimit
from onegov.form.validators import WhitelistedMimeType
from onegov.gis import CoordinatesField
from onegov.org import _
from sedate import replace_timezone, to_timezone
from wtforms import RadioField, StringField, TextAreaField, validators
from wtforms.fields.html5 import DateField, EmailField
from wtforms_components import TimeField


TAGS = [(tag, tag) for tag in (
    _("Exhibition"),
    _("Gastronomy"),
    _("Health"),
    _("Cinema"),
    _("Congress"),
    _("Concert"),
    _("Art"),
    _("Literature"),
    _("Market"),
    _("Music"),
    _("Party"),
    _("Politics"),
    _("Religion"),
    _("Sports"),
    _("Dancing"),
    _("Theater"),
    _("Meetup"),
    _("Talk"),
)]

WEEKDAYS = (
    ("MO", _("Mo")),
    ("TU", _("Tu")),
    ("WE", _("We")),
    ("TH", _("Th")),
    ("FR", _("Fr")),
    ("SA", _("Sa")),
    ("SU", _("Su")),
)


class EventForm(Form):
    """ Defines the form for all events. """

    timezone = 'Europe/Zurich'

    email = EmailField(
        label=_("Submitter"),
        description="max.muster@example.org",
        validators=[validators.InputRequired(), validators.Email()]
    )

    title = StringField(
        label=_("Title"),
        description=_("Concerto in the castle garden"),
        validators=[validators.InputRequired()]
    )

    description = TextAreaField(
        label=_("Description"),
        description=_("Enjoy a concerto in the castle garden."),
        render_kw={'rows': 12}
    )

    image = UploadFileWithORMSupport(
        label=_("Image"),
        description="Foobar",
        file_class=EventFile,
        validators=[
            validators.Optional(),
            WhitelistedMimeType({
                'image/gif',
                'image/jpeg',
                'image/png'
            }),
            FileSizeLimit(1 * 1024 * 1024)
        ]
    )

    location = StringField(
        label=_("Venue"),
        description=_("Castle garden"),
        validators=[validators.InputRequired()]
    )

    organizer = StringField(
        label=_("Organizer"),
        description=_("Music society"),
        validators=[validators.InputRequired()]
    )

    coordinates = CoordinatesField(
        label=_("Coordinates"),
        description=_("The marker can be moved by dragging it with the mouse"),
        render_kw={'data-map-type': 'marker'}
    )

    tags = MultiCheckboxField(
        label=_("Tags"),
        choices=TAGS,
    )

    start_date = DateField(
        label=_("Date"),
        validators=[validators.InputRequired()],
        default=date.today
    )

    start_time = TimeField(
        label=_("From"),
        description="18:00",
        validators=[validators.InputRequired()]
    )

    end_time = TimeField(
        label=_("To"),
        description="19:15",
        validators=[validators.InputRequired()]
    )

    repeat = RadioField(
        label=_("Repeat"),
        default='without',
        choices=(
            ('without', _("Without")),
            ('weekly', _("Weekly")),
            ('dates', _("On additional dates"))
        )
    )

    weekly = MultiCheckboxField(
        label=_("Repeats itself every"),
        choices=WEEKDAYS,
        render_kw={'prefix_label': False, 'class_': 'oneline-checkboxes'},
        depends_on=('repeat', 'weekly')
    )

    end_date = DateField(
        label=_("Until date"),
        validators=[validators.Optional()],
        depends_on=('repeat', 'weekly')
    )

    dates = TextAreaField(
        label=_("Dates"),
        depends_on=('repeat', 'dates'),
        render_kw={'class_': 'many many-dates'},
    )

    @property
    def start(self):
        return replace_timezone(
            datetime(
                self.start_date.data.year,
                self.start_date.data.month,
                self.start_date.data.day,
                self.start_time.data.hour,
                self.start_time.data.minute
            ),
            self.timezone
        )

    @property
    def end(self):
        end_date = self.start_date.data

        if self.end_time.data <= self.start_time.data:
            end_date += timedelta(days=1)

        return replace_timezone(
            datetime(
                end_date.year,
                end_date.month,
                end_date.day,
                self.end_time.data.hour,
                self.end_time.data.minute
            ),
            self.timezone
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.date_errors = {}

    def on_request(self):
        self.request.include('common')
        self.request.include('many')

        if not self.dates.data:
            self.dates.data = self.dates_to_json(None)

    def validate(self):
        """ Make sure a valid RRULE can be generated with the given fields.

        Might be better to group weekly and end_date in an enclosure,
        see See http://wtforms.readthedocs.org/en/latest/fields.html
        #field-enclosures.

        """
        result = super().validate()

        if self.start_date.data and self.end_date.data:
            if self.start_date.data > self.end_date.data:
                message = _("The end date must be later than the start date.")
                self.end_date.errors.append(message)
                result = False

        if self.repeat.data == 'weekly':
            if self.weekly.data and self.start_date.data:
                weekday = WEEKDAYS[self.start_date.data.weekday()][0]
                if weekday not in self.weekly.data:
                    message = _(
                        "The weekday of the start date must be selected.")
                    self.weekly.errors.append(message)
                    result = False

            if self.weekly.data and not self.end_date.data:
                message = _(
                    "Please set and end date if the event is recurring.")
                self.end_date.errors.append(message)
                result = False

            if self.end_date.data and not self.weekly.data:
                message = _(
                    "Please select a weekday if the event is recurring.")
                self.weekly.errors.append(message)
                result = False

        return result

    def populate_obj(self, model):
        """ Stores the form values on the model. """

        super().populate_obj(model, exclude={
            'start_date',
            'start_time',
            'end_date',
            'end_time',
            'weekly',
            'email',
        })

        # clear the recurrence to avoid updating all occurrences too much
        model.recurrence = ''

        model.meta = model.meta or {}
        model.meta['submitter_email'] = self.email.data
        model.timezone = self.timezone
        model.start = self.start
        model.end = self.end

        if self.repeat.data == 'without':
            self.recurrence = None
        elif self.repeat.data == 'weekly':
            if self.weekly.data and self.end_date.data:
                until_date = to_timezone(
                    replace_timezone(
                        datetime(
                            self.end_date.data.year,
                            self.end_date.data.month,
                            self.end_date.data.day,
                            self.end_time.data.hour,
                            self.end_time.data.minute
                        ), model.timezone
                    ), 'UTC'
                )
                model.recurrence = (
                    "RRULE:FREQ=WEEKLY;WKST=MO;BYDAY={0};UNTIL={1}".format(
                        ','.join(self.weekly.data),
                        until_date.strftime('%Y%m%dT%H%M%SZ')
                    )
                )
        elif self.repeat.data == 'dates':
            model.recurrence = '\n'.join(
                f"RDATE:{d.strftime('%Y%m%dT000000Z')}"
                for d in self.parsed_dates
            )
        else:
            raise NotImplementedError

    def process_obj(self, model):
        """ Stores the page values on the form. """

        super().process_obj(model)

        self.start_time.data = model.localized_start.time()
        self.end_time.data = model.localized_end.time()
        self.start_date.data = model.localized_start.date()

        if model.recurrence:
            last_occurrence = model.occurrence_dates(localize=True)[-1]
            self.end_date.data = last_occurrence.date()

            rule = rrulestr(model.recurrence)

            if getattr(rule, '_freq', None) == rrule.WEEKLY:
                self.repeat.data = 'weekly'
                self.dates.data = self.dates_to_json(None)
                self.weekly.data = [
                    WEEKDAYS[day][0] for day in rule._byweekday
                ]
            else:
                self.repeat.data = 'dates'
                self.dates.data = self.dates_to_json(tuple(rule))
        elif self.errors.get('weekly'):
            self.dates.data = self.dates_to_json(None)
            self.repeat.data = 'weekly'
        else:
            self.dates.data = self.dates_to_json(None)
            self.repeat.data = 'without'

        if model.meta:
            self.email.data = model.meta.get('submitter_email')

    @cached_property
    def parsed_dates(self):
        return self.json_to_dates(self.dates.data)

    def json_to_dates(self, text=None):
        result = []

        for value in json.loads(text or '{}').get('values', []):
            result.append(date(*map(int, value['date'].split('-'))))

        return result

    def dates_to_json(self, dates=None):
        dates = dates or []

        return json.dumps({
            'labels': {
                'date': self.request.translate(_("Date")),
                'add': self.request.translate(_("Add")),
                'remove': self.request.translate(_("Remove")),
            },
            'values': [
                {
                    'date': d.strftime('%Y-%m-%d'),
                    'error': self.date_errors.get(ix, "")
                } for ix, d in enumerate(dates)
            ]
        })
